from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from .performance_test import InfrahubPerformanceTest

if TYPE_CHECKING:
    import _pytest


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup("infrahub-performance-test")

    group.addoption(
        "--performance-result-address",
        action="store",
        dest="infrahub_performance_test_result_address",
        default="https://webhook.site/25802839-4b34-433e-9dc4-59623cd73c80",
        metavar="INFRAHUB_PERFORMANCE_TEST_RESULT_ADDRESS",
        help="Address to send the results of the performance test (default: %(default)s)",
    )
    group.addoption(
        "--performance-backup",
        action="store",
        dest="infrahub_performance_backup",
        default="neo4j_database.db",
        metavar="INFRAHUB_PERFORMANCE_BACKUP",
        help="Name of the backup to use (default: %(default)s)",
    )
    group.addoption(
        "--performance-backup-location",
        action="store",
        dest="infrahub_performance_backup_location",
        metavar="INFRAHUB_PERFORMANCE_BACKUP_LOCATION",
        help="Location of the backup to use (default: %(default)s)",
    )
    group.addoption(
        "--performance-create-backup",
        action="store_true",
        default=False,
        dest="infrahub_performance_create_backup",
        help="Generate a backup of the database (default: %(default)s)",
    )
    group.addoption(
        "--performance-use-backup",
        action="store_true",
        default=False,
        dest="infrahub_performance_use_backup",
        help="Use a backup of the database and skip all tests that creates data (default: %(default)s)",
    )
    group.addoption(
        "--performance-report",
        action="store_true",
        dest="infrahub_performance_report",
        default=False,
        help="Display performance report at the end of the test run",
    )


def pytest_collection_modifyitems(config: pytest.Config, items: list[pytest.Item]) -> None:
    use_backup_skip = pytest.mark.skip(reason="load data from backup")
    no_create_backup_skip = pytest.mark.skip(reason="no need to create a backup")
    no_load_backup_skip = pytest.mark.skip(reason="no need to load a backup")

    use_backup = config.getoption("infrahub_performance_use_backup")
    create_backup = config.getoption("infrahub_performance_create_backup")

    for item in items:
        has_create_backup_marker = bool(
            [True for marker in item.own_markers if marker.name == "performance_create_backup"]
        )
        has_load_backup_marker = bool([True for marker in item.own_markers if marker.name == "performance_load_backup"])
        has_load_data_marker = bool([True for marker in item.own_markers if marker.name == "performance_load_data"])

        if has_create_backup_marker and not create_backup:
            item.add_marker(no_create_backup_skip)
        if has_load_backup_marker and not use_backup:
            item.add_marker(no_load_backup_skip)
        if has_load_data_marker and use_backup:
            item.add_marker(use_backup_skip)


def pytest_sessionstart(session: pytest.Session) -> None:
    session.infrahub_performance_test = InfrahubPerformanceTest(
        results_url=session.config.getoption("infrahub_performance_test_result_address")
    )


def pytest_sessionfinish(session: pytest.Session, exitstatus: int) -> None:  # noqa: ARG001
    """whole test run finishes."""
    if not session.config.getoption("infrahub_performance_report"):
        return

    session.infrahub_performance_test.finalize(session=session)


def pytest_runtest_teardown(item: pytest.Item, nextitem: pytest.Item | None) -> None:  # noqa: ARG001
    """Fetch metrics at each test teardown because there's no better hook...
    pytest_sessionfinish() is executed after fixtures has been finalized and pytest_fixture_post_finalizer() is too late"""
    if not item.config.getoption("infrahub_performance_report"):
        return

    item.session.infrahub_performance_test.fetch_metrics()


def pytest_configure(config: pytest.Config) -> None:
    if config.getoption("infrahub_performance_use_backup") and config.getoption("infrahub_performance_create_backup"):
        raise pytest.UsageError("--performance-use-backup and --performance-create-backup are mutually exclusive")

    config.addinivalue_line("markers", "performance_create_backup: Create a backup of the database")
    config.addinivalue_line("markers", "performance_load_backup: Load the backup of the database")
    config.addinivalue_line("markers", "performance_load_data: Load initial data into the database")


def pytest_terminal_summary(
    terminalreporter: _pytest.terminal.TerminalReporter,
    exitstatus: int,  # noqa: ARG001
    config: pytest.Config,
) -> None:
    if not config.getoption("infrahub_performance_report"):
        return

    performance_test = terminalreporter._session.infrahub_performance_test

    report = [
        f"{measurement.name} ({measurement.context}): {measurement.value} {measurement.unit.value}"
        for measurement in performance_test.measurements
    ]
    terminalreporter.write("\n" + "\n".join(report) + "\n")


@pytest.fixture(scope="session")
def perf_test(request: pytest.FixtureRequest) -> InfrahubPerformanceTest:
    return request.session.infrahub_performance_test



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Self

from infrahub.core import registry
from infrahub.core.path import SchemaPath  # noqa: TC001
from infrahub.core.query import Query  # noqa: TC001
from infrahub.core.schema import (
    AttributeSchema,
    GenericSchema,
    NodeSchema,
    RelationshipSchema,
    SchemaRoot,
    internal_schema,
)

from .query import MigrationQuery  # noqa: TC001

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class MigrationResult(BaseModel):
    errors: list[str] = Field(default_factory=list)
    nbr_migrations_executed: int = 0

    @property
    def success(self) -> bool:
        if not self.errors:
            return True

        return False


class SchemaMigration(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(..., description="Name of the migration")
    queries: Sequence[type[MigrationQuery]] = Field(..., description="List of queries to execute for this migration")

    new_node_schema: NodeSchema | GenericSchema | None = None
    previous_node_schema: NodeSchema | GenericSchema | None = None
    schema_path: SchemaPath

    async def execute(self, db: InfrahubDatabase, branch: Branch, at: Timestamp | str | None = None) -> MigrationResult:
        async with db.start_transaction() as ts:
            result = MigrationResult()

            for migration_query in self.queries:
                try:
                    query = await migration_query.init(db=ts, branch=branch, at=at, migration=self)
                    await query.execute(db=ts)
                    result.nbr_migrations_executed += query.get_nbr_migrations_executed()
                except Exception as exc:
                    result.errors.append(str(exc))
                    return result

        return result

    @property
    def new_schema(self) -> NodeSchema | GenericSchema:
        if self.new_node_schema:
            return self.new_node_schema
        raise ValueError("new_node_schema hasn't been initialized")

    @property
    def previous_schema(self) -> NodeSchema | GenericSchema:
        if self.previous_node_schema:
            return self.previous_node_schema
        raise ValueError("previous_node_schema hasn't been initialized")


class AttributeSchemaMigration(SchemaMigration):
    @property
    def new_attribute_schema(self) -> AttributeSchema:
        if not self.schema_path.field_name:
            raise ValueError("field_name is not defined")
        return self.new_schema.get_attribute(name=self.schema_path.field_name)

    @property
    def previous_attribute_schema(self) -> AttributeSchema:
        if not self.schema_path.field_name:
            raise ValueError("field_name is not defined")
        return self.previous_schema.get_attribute(name=self.schema_path.field_name)


class RelationshipSchemaMigration(SchemaMigration):
    @property
    def new_relationship_schema(self) -> RelationshipSchema:
        if not self.schema_path.field_name:
            raise ValueError("field_name is not defined")
        return self.new_schema.get_relationship(name=self.schema_path.field_name)

    @property
    def previous_relationship_schema(self) -> RelationshipSchema:
        if not self.schema_path.field_name or not self.previous_node_schema:
            raise ValueError("field_name is not defined")
        return self.previous_schema.get_relationship(name=self.schema_path.field_name)


class GraphMigration(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(..., description="Name of the migration")
    queries: Sequence[type[Query]] = Field(..., description="List of queries to execute for this migration")
    minimum_version: int = Field(..., description="Minimum version of the graph to execute this migration")

    @classmethod
    def init(cls, **kwargs: dict[str, Any]) -> Self:
        return cls(**kwargs)  # type: ignore[arg-type]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        raise NotImplementedError

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        async with db.start_transaction() as ts:
            return await self.do_execute(db=ts)

    async def do_execute(self, db: InfrahubDatabase) -> MigrationResult:
        result = MigrationResult()
        for migration_query in self.queries:
            try:
                query = await migration_query.init(db=db)
                await query.execute(db=db)
            except Exception as exc:
                result.errors.append(str(exc))
                return result

        return result


class InternalSchemaMigration(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(..., description="Name of the migration")
    migrations: Sequence[SchemaMigration] = Field(..., description="")
    minimum_version: int = Field(..., description="Minimum version of the graph to execute this migration")

    @staticmethod
    def get_internal_schema() -> SchemaBranch:
        from infrahub.core.schema.schema_branch import SchemaBranch

        # load the internal schema from
        schema = SchemaRoot(**internal_schema)
        schema_branch = SchemaBranch(cache={}, name="default_branch")
        schema_branch.load_schema(schema=schema)
        schema_branch.process()

        return schema_branch

    @classmethod
    def init(cls, **kwargs: dict[str, Any]) -> Self:
        return cls(**kwargs)  # type: ignore[arg-type]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        raise NotImplementedError

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        result = MigrationResult()

        default_branch = registry.get_branch_from_registry()

        for migration in self.migrations:
            try:
                execution_result = await migration.execute(db=db, branch=default_branch)
                result.errors.extend(execution_result.errors)
            except Exception as exc:
                result.errors.append(str(exc))
                return result

        return result


class ArbitraryMigration(BaseModel):
    name: str = Field(..., description="Name of the migration")
    minimum_version: int = Field(..., description="Minimum version of the graph to execute this migration")

    @classmethod
    def init(cls, **kwargs: dict[str, Any]) -> Self:
        return cls(**kwargs)  # type: ignore[arg-type]

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:
        raise NotImplementedError()

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        raise NotImplementedError()



from typing import ClassVar

from pydantic import Field

from infrahub.message_bus import InfrahubMessage
from infrahub.message_bus.messages.refresh_registry_branches import RefreshRegistryBranches

from .constants import EVENT_NAMESPACE
from .models import InfrahubEvent


class SchemaUpdatedEvent(InfrahubEvent):
    """Event generated when the schema within a branch has been updated."""

    event_name: ClassVar[str] = f"{EVENT_NAMESPACE}.schema.updated"

    branch_name: str = Field(..., description="The name of the branch")
    schema_hash: str = Field(..., description="Schema hash after the update")

    # NOTE
    # We could add to the payload
    # - Hash before and after the change
    # - List of nodes and generics that have been modified
    # - Diff of the change

    # NOTE 2
    # Should schema_update be a branch event ?
    # if feels like the main resource should be the branch

    def get_resource(self) -> dict[str, str]:
        return {
            "prefect.resource.id": f"infrahub.schema_branch.{self.branch_name}",
            "infrahub.branch.name": self.branch_name,
            "infrahub.branch.schema_hash": self.schema_hash,
        }

    def get_messages(self) -> list[InfrahubMessage]:
        return [
            RefreshRegistryBranches(),
            # EventSchemaUpdate(
            #     branch=self.branch,
            #     meta=self.get_message_meta(),
            # )
        ]



from typing import AsyncGenerator

from infrahub.core.query import QueryType
from infrahub.database import InfrahubDatabase

from .models import VertexToDelete


class PatchPlanVertexDeleter:
    def __init__(self, db: InfrahubDatabase, batch_size_limit: int = 1000) -> None:
        self.db = db
        self.batch_size_limit = batch_size_limit

    async def _run_delete_query(self, ids_to_delete: list[str]) -> set[str]:
        query = """
MATCH (n)
WHERE %(id_func_name)s(n) IN $ids_to_delete
DETACH DELETE n
RETURN %(id_func_name)s(n) AS deleted_id
        """ % {"id_func_name": self.db.get_id_function_name()}
        results = await self.db.execute_query(
            query=query, params={"ids_to_delete": ids_to_delete}, type=QueryType.WRITE
        )
        deleted_ids: set[str] = set()
        for result in results:
            deleted_id = result.get("deleted_id")
            deleted_ids.add(deleted_id)
        return deleted_ids

    async def execute(self, vertices_to_delete: list[VertexToDelete]) -> AsyncGenerator[set[str], None]:
        for i in range(0, len(vertices_to_delete), self.batch_size_limit):
            ids_to_delete = [v.db_id for v in vertices_to_delete[i : i + self.batch_size_limit]]
            yield await self._run_delete_query(ids_to_delete=ids_to_delete)



COLOR_SELECTION = [
    "#ed6a5a",
    "#f4f1bb",
    "#9bc1bc",
    "#5ca4a9",
    "#e6ebe0",
    "#52489c",
    "#4062bb",
    "#59c3c3",
    "#56638a",
    "#5d737e",
    "#55505c",
    "#592941",
    "#498467",
    "#e5d0e3",
    "#748cab",
    "#018e42",
    "#50808e",
    "#7b2d26",
    "#9799ca",
    "#51bbfe",
]


def select_color(existing: list[str]) -> str:
    """Select a color from a predefined list without including anything from a list of existing colors."""
    for color in COLOR_SELECTION:
        if color not in existing:
            existing.append(color)
            return color

    return "#ff0000"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Self, cast
from uuid import UUID

from pydantic import BaseModel, Field, PrivateAttr, computed_field, field_validator, model_validator

from infrahub.core.constants import NULL_VALUE, DiffAction, RelationshipCardinality, RelationshipKind

if TYPE_CHECKING:
    from infrahub.core.attribute import BaseAttribute
    from infrahub.core.branch import Branch
    from infrahub.core.manager import RelationshipSchema
    from infrahub.core.query.relationship import RelationshipPeerData
    from infrahub.core.relationship.model import Relationship
    from infrahub.core.schema import MainSchemaTypes
    from infrahub.core.schema.schema_branch import SchemaBranch
    from infrahub.database import InfrahubDatabase


class PropertyChangelog(BaseModel):
    name: str = Field(..., description="The name of the property")
    value: str | bool | None = Field(..., description="The updated or current value of the property")
    value_previous: str | bool | None = Field(
        ...,
        description="The previous value of the property, a `null` value indicates that the property didn't previously have a value",
    )

    @computed_field
    def value_type(self) -> str:
        """The value_type of the property, used to help external systems"""
        if isinstance(self.value, str):
            return "Text"

        return "Boolean"

    @computed_field
    def value_update_status(self) -> DiffAction:
        """Indicate how the value was changed during this update"""
        if self.value == self.value_previous:
            return DiffAction.UNCHANGED
        if self.value_previous is not None and self.value is None:
            return DiffAction.REMOVED
        if self.value_previous is None and self.value is not None:
            return DiffAction.ADDED

        return DiffAction.UPDATED


class AttributeChangelog(BaseModel):
    name: str = Field(..., description="The name of the attribute")
    value: Any = Field(default=None, description="The current value of the attribute")
    value_previous: Any = Field(default=None, description="The previous value of the attribute")
    properties: dict[str, PropertyChangelog] = Field(
        default_factory=dict, description="The properties that were updated during this update"
    )
    kind: str = Field(..., description="The attribute kind")

    @computed_field
    def value_update_status(self) -> DiffAction:
        """Indicate how the peer was changed during this update"""
        if self.value == self.value_previous:
            return DiffAction.UNCHANGED
        if self.value_previous is not None and self.value is None:
            return DiffAction.REMOVED
        if self.value_previous is None and self.value is not None:
            return DiffAction.ADDED

        return DiffAction.UPDATED

    def add_property(self, name: str, value_current: bool | str | None, value_previous: bool | str | None) -> None:
        self.properties[name] = PropertyChangelog(name=name, value=value_current, value_previous=value_previous)

    @property
    def has_updates(self) -> bool:
        if self.value_update_status != DiffAction.UNCHANGED or self.properties:
            return True
        return False

    def set_value(self, value: Any) -> None:
        if isinstance(value, str) and value == NULL_VALUE:
            self.value = None
            return
        self.value = value

    def set_value_previous(self, value: Any) -> None:
        if isinstance(value, str) and value == NULL_VALUE:
            self.value_previous = None
            return
        self.value_previous = value

    @field_validator("value", "value_previous")
    @classmethod
    def convert_null_values(cls, value: Any) -> Any:
        if isinstance(value, str) and value == NULL_VALUE:
            return None
        return value

    @model_validator(mode="after")
    def filter_sensitive(self) -> Self:
        if self.kind in ["HashedPassword", "Password"]:
            if self.value is not None:
                self.value = "***"
            if self.value_previous is not None:
                self.value_previous = "***"

        return self


class RelationshipCardinalityOneChangelog(BaseModel):
    name: str = Field(..., description="The name of the relationship")
    peer_id_previous: str | None = Field(default=None, description="The previous peer of this relationship")
    peer_kind_previous: str | None = Field(default=None, description="The node kind of the previous peer")
    peer_id: str | None = Field(default=None, description="The current peer of this relationship")
    peer_kind: str | None = Field(default=None, description="The node kind of the current peer")
    properties: dict[str, PropertyChangelog] = Field(
        default_factory=dict, description="Changes to properties of this relationship if any were made"
    )
    _parent: ChangelogRelatedNode | None = PrivateAttr(default=None)

    @property
    def parent(self) -> ChangelogRelatedNode | None:
        return self._parent

    @computed_field
    def cardinality(self) -> str:
        return "one"

    @computed_field
    def peer_status(self) -> DiffAction:
        """Indicate how the peer was changed during this update"""
        if self.peer_id_previous == self.peer_id:
            return DiffAction.UNCHANGED
        if self.peer_id_previous and not self.peer_id:
            return DiffAction.REMOVED
        if self.peer_id and not self.peer_id_previous:
            return DiffAction.ADDED
        return DiffAction.UPDATED

    def add_property(self, name: str, value_current: bool | str | None, value_previous: bool | str | None) -> None:
        self.properties[name] = PropertyChangelog(name=name, value=value_current, value_previous=value_previous)

    def set_parent(self, parent_id: str, parent_kind: str) -> None:
        self._parent = ChangelogRelatedNode(node_id=parent_id, node_kind=parent_kind)

    def set_parent_from_relationship(self, rel_kind: RelationshipKind) -> None:
        if rel_kind == RelationshipKind.PARENT:
            if (
                self.peer_status in [DiffAction.ADDED, DiffAction.UNCHANGED, DiffAction.UPDATED]
                and self.peer_id
                and self.peer_kind
            ):
                self._parent = ChangelogRelatedNode(node_id=self.peer_id, node_kind=self.peer_kind)
            elif self.peer_id_previous and self.peer_kind_previous:
                self._parent = ChangelogRelatedNode(node_id=self.peer_id_previous, node_kind=self.peer_kind_previous)

    @property
    def is_empty(self) -> bool:
        return self.peer_status == DiffAction.UNCHANGED and not self.properties


class RelationshipPeerChangelog(BaseModel):
    peer_id: str = Field(..., description="The ID of the peer")
    peer_kind: str = Field(..., description="The node kind of the peer")
    peer_status: DiffAction = Field(
        ..., description="Indicate how the relationship to this peer was changed in this update"
    )
    properties: dict[str, PropertyChangelog] = Field(
        default_factory=dict, description="Changes to properties of this relationship if any were made"
    )

    def add_property(self, name: str, value_current: bool | str | None, value_previous: bool | str | None) -> None:
        self.properties[name] = PropertyChangelog(name=name, value=value_current, value_previous=value_previous)


class RelationshipCardinalityManyChangelog(BaseModel):
    name: str
    peers: list[RelationshipPeerChangelog] = Field(default_factory=list)

    @computed_field
    def cardinality(self) -> str:
        return "many"

    def add_new_peer(self, relationship: Relationship) -> None:
        properties: dict[str, PropertyChangelog] = {}
        properties["is_protected"] = PropertyChangelog(
            name="is_protected", value=relationship.is_protected, value_previous=None
        )
        properties["is_visible"] = PropertyChangelog(
            name="is_visible", value=relationship.is_protected, value_previous=None
        )
        if owner := getattr(relationship, "owner_id", None):
            properties["owner"] = PropertyChangelog(name="owner", value=owner, value_previous=None)
        if source := getattr(relationship, "source_id", None):
            properties["source"] = PropertyChangelog(name="source_id", value=source, value_previous=None)

        self.peers.append(
            RelationshipPeerChangelog(
                peer_id=relationship.get_peer_id(),
                peer_kind=relationship.get_peer_kind(),
                peer_status=DiffAction.ADDED,
                properties=properties,
            )
        )

    def remove_peer(self, peer_id: str, peer_kind: str) -> None:
        self.peers.append(
            RelationshipPeerChangelog(
                peer_id=peer_id,
                peer_kind=peer_kind,
                peer_status=DiffAction.REMOVED,
            )
        )

    @property
    def is_empty(self) -> bool:
        return not self.peers


class ChangelogRelatedNode(BaseModel):
    node_id: str
    node_kind: str


class NodeChangelog(BaseModel):
    """Emitted when a node is updated"""

    node_id: str
    node_kind: str
    display_label: str

    attributes: dict[str, AttributeChangelog] = Field(default_factory=dict)
    relationships: dict[str, RelationshipCardinalityOneChangelog | RelationshipCardinalityManyChangelog] = Field(
        default_factory=dict
    )

    _parent: ChangelogRelatedNode | None = PrivateAttr(default=None)

    @property
    def parent(self) -> ChangelogRelatedNode | None:
        return self._parent

    @property
    def updated_fields(self) -> list[str]:
        """Return a list of update fields i.e. attributes and relationships"""
        return list(self.relationships.keys()) + list(self.attributes.keys())

    @property
    def has_changes(self) -> bool:
        return len(self.updated_fields) > 0

    @property
    def root_node_id(self) -> str:
        """Return the top level node_id"""
        if self.parent:
            return self.parent.node_id
        return self.node_id

    def add_parent(self, parent: ChangelogRelatedNode) -> None:
        self._parent = parent

    def add_parent_from_relationship(self, parent: Relationship) -> None:
        self._parent = ChangelogRelatedNode(node_id=parent.get_peer_id(), node_kind=parent.get_peer_kind())

    def create_relationship(self, relationship: Relationship) -> None:
        if relationship.schema.cardinality == RelationshipCardinality.ONE:
            peer_id = relationship.get_peer_id()
            peer_kind = relationship.get_peer_kind()
            if relationship.schema.kind == RelationshipKind.PARENT:
                self._parent = ChangelogRelatedNode(node_id=peer_id, node_kind=peer_kind)
            changelog_relationship = RelationshipCardinalityOneChangelog(
                name=relationship.schema.name,
                peer_id=peer_id,
                peer_kind=peer_kind,
            )
            if source_id := getattr(relationship, "source_id", None):
                changelog_relationship.add_property(name="source", value_current=source_id, value_previous=None)
            if owner_id := getattr(relationship, "owner_id", None):
                changelog_relationship.add_property(name="owner", value_current=owner_id, value_previous=None)
            changelog_relationship.add_property(
                name="is_protected", value_current=relationship.is_protected, value_previous=None
            )
            changelog_relationship.add_property(
                name="is_visible", value_current=relationship.is_visible, value_previous=None
            )
            self.relationships[changelog_relationship.name] = changelog_relationship
        elif relationship.schema.cardinality == RelationshipCardinality.MANY:
            if relationship.schema.name not in self.relationships:
                self.relationships[relationship.schema.name] = RelationshipCardinalityManyChangelog(
                    name=relationship.schema.name
                )
            relationship_container = cast(
                RelationshipCardinalityManyChangelog, self.relationships[relationship.schema.name]
            )

            relationship_container.add_new_peer(relationship=relationship)

    def delete_relationship(self, relationship: Relationship) -> None:
        if relationship.schema.cardinality == RelationshipCardinality.ONE:
            peer_id = relationship.get_peer_id()
            peer_kind = relationship.get_peer_kind()
            changelog_relationship = RelationshipCardinalityOneChangelog(
                name=relationship.schema.name,
                peer_id_previous=peer_id,
                peer_kind_previous=peer_kind,
            )
            self.relationships[changelog_relationship.name] = changelog_relationship
        elif relationship.schema.cardinality == RelationshipCardinality.MANY:
            if relationship.schema.name not in self.relationships:
                self.relationships[relationship.schema.name] = RelationshipCardinalityManyChangelog(
                    name=relationship.schema.name
                )
            relationship_container = cast(
                RelationshipCardinalityManyChangelog, self.relationships[relationship.schema.name]
            )
            relationship_container.remove_peer(
                peer_id=relationship.get_peer_id(), peer_kind=relationship.get_peer_kind()
            )

    def add_attribute(self, attribute: AttributeChangelog) -> None:
        if attribute.has_updates:
            self.attributes[attribute.name] = attribute

    def add_relationship(
        self, relationship_changelog: RelationshipCardinalityOneChangelog | RelationshipCardinalityManyChangelog
    ) -> None:
        if isinstance(relationship_changelog, RelationshipCardinalityOneChangelog) and relationship_changelog.parent:
            self.add_parent(parent=relationship_changelog.parent)
        if relationship_changelog.is_empty:
            return

        self.relationships[relationship_changelog.name] = relationship_changelog

    def create_attribute(self, attribute: BaseAttribute) -> None:
        changelog_attribute = AttributeChangelog(
            name=attribute.name, value=attribute.value, value_previous=None, kind=attribute.schema.kind
        )
        if source_id := getattr(attribute, "source_id", None):
            changelog_attribute.add_property(name="source", value_current=source_id, value_previous=None)
        if owner_id := getattr(attribute, "owner_id", None):
            changelog_attribute.add_property(name="owner", value_current=owner_id, value_previous=None)
        changelog_attribute.add_property(name="is_protected", value_current=attribute.is_protected, value_previous=None)
        changelog_attribute.add_property(name="is_visible", value_current=attribute.is_visible, value_previous=None)
        self.attributes[changelog_attribute.name] = changelog_attribute

    def get_related_nodes(self) -> list[ChangelogRelatedNode]:
        related_nodes: dict[str, ChangelogRelatedNode] = {}
        for relationship in self.relationships.values():
            if isinstance(relationship, RelationshipCardinalityOneChangelog):
                if relationship.peer_id and relationship.peer_kind:
                    related_nodes[relationship.peer_id] = ChangelogRelatedNode(
                        node_id=relationship.peer_id, node_kind=relationship.peer_kind
                    )
                if relationship.peer_id_previous and relationship.peer_kind_previous:
                    related_nodes[relationship.peer_id_previous] = ChangelogRelatedNode(
                        node_id=relationship.peer_id_previous, node_kind=relationship.peer_kind_previous
                    )
            elif isinstance(relationship, RelationshipCardinalityManyChangelog):
                for peer in relationship.peers:
                    related_nodes[peer.peer_id] = ChangelogRelatedNode(node_id=peer.peer_id, node_kind=peer.peer_kind)

        if self.parent:
            related_nodes[self.parent.node_id] = self.parent

        return list(related_nodes.values())


class ChangelogRelationshipMapper:
    def __init__(self, schema: RelationshipSchema) -> None:
        self.schema = schema
        self._cardinality_one_relationship: RelationshipCardinalityOneChangelog | None = None
        self._cardinality_many_relationship: RelationshipCardinalityManyChangelog | None = None

    @property
    def cardinality_one_relationship(self) -> RelationshipCardinalityOneChangelog:
        if not self._cardinality_one_relationship:
            self._cardinality_one_relationship = RelationshipCardinalityOneChangelog(name=self.schema.name)

        return self._cardinality_one_relationship

    @property
    def cardinality_many_relationship(self) -> RelationshipCardinalityManyChangelog:
        if not self._cardinality_many_relationship:
            self._cardinality_many_relationship = RelationshipCardinalityManyChangelog(name=self.schema.name)

        return self._cardinality_many_relationship

    def remove_peer(self, peer_data: RelationshipPeerData) -> None:
        if self.schema.cardinality == RelationshipCardinality.ONE:
            self.cardinality_one_relationship.peer_id_previous = str(peer_data.peer_id)
            self.cardinality_one_relationship.peer_kind_previous = peer_data.peer_kind
        elif self.schema.cardinality == RelationshipCardinality.MANY:
            self.cardinality_many_relationship.remove_peer(
                peer_id=str(peer_data.peer_id), peer_kind=peer_data.peer_kind
            )

    def _set_cardinality_one_peer(self, relationship: Relationship) -> None:
        self.cardinality_one_relationship.peer_id = relationship.peer_id
        self.cardinality_one_relationship.peer_kind = relationship.get_peer_kind()
        self.cardinality_one_relationship.set_parent_from_relationship(rel_kind=relationship.schema.kind)

    def add_parent_from_relationship(self, relationship: Relationship) -> None:
        if self.schema.cardinality == RelationshipCardinality.ONE:
            self.cardinality_one_relationship.set_parent(
                parent_id=relationship.get_peer_id(), parent_kind=relationship.get_peer_kind()
            )

    def add_peer_from_relationship(self, relationship: Relationship) -> None:
        if self.schema.cardinality == RelationshipCardinality.ONE:
            self._set_cardinality_one_peer(relationship=relationship)
        elif self.schema.cardinality == RelationshipCardinality.MANY:
            self.cardinality_many_relationship.add_new_peer(relationship=relationship)

    def add_updated_relationship(
        self, relationship: Relationship, old_data: RelationshipPeerData, properties_to_update: list[str]
    ) -> None:
        if self.schema.cardinality == RelationshipCardinality.ONE:
            self._set_cardinality_one_peer(relationship=relationship)
            self.cardinality_one_relationship.peer_id_previous = self.cardinality_one_relationship.peer_id
            self.cardinality_one_relationship.peer_kind_previous = self.cardinality_one_relationship.peer_kind
            for property_to_update in properties_to_update:
                previous_property = old_data.properties.get(property_to_update)
                previous_value: str | bool | None = None
                if previous_property:
                    if isinstance(previous_property.value, UUID):
                        previous_value = str(previous_property.value)
                    else:
                        previous_value = previous_property.value
                property_name = (
                    property_to_update if property_to_update not in ["source", "owner"] else f"{property_to_update}_id"
                )
                self.cardinality_one_relationship.add_property(
                    name=property_to_update,
                    value_current=getattr(relationship, property_name),
                    value_previous=previous_value,
                )
            self.cardinality_one_relationship.set_parent_from_relationship(rel_kind=relationship.schema.kind)

    def delete_relationship(self, peer_id: str, peer_kind: str, rel_schema: RelationshipSchema) -> None:
        if self.schema.cardinality == RelationshipCardinality.ONE:
            self.cardinality_one_relationship.peer_id_previous = peer_id
            self.cardinality_one_relationship.peer_kind_previous = peer_kind
            self.cardinality_one_relationship.set_parent_from_relationship(rel_kind=rel_schema.kind)

        elif self.schema.cardinality == RelationshipCardinality.MANY:
            self.cardinality_many_relationship.remove_peer(peer_id=peer_id, peer_kind=peer_kind)

    @property
    def changelog(self) -> RelationshipCardinalityOneChangelog | RelationshipCardinalityManyChangelog:
        match self.schema.cardinality:
            case RelationshipCardinality.ONE:
                return self.cardinality_one_relationship
            case RelationshipCardinality.MANY:
                return self.cardinality_many_relationship


class RelationshipChangelogGetter:
    def __init__(self, db: InfrahubDatabase, branch: Branch) -> None:
        self._db = db
        self._branch = branch

    async def get_changelogs(self, primary_changelog: NodeChangelog) -> list[NodeChangelog]:
        """Return secondary changelogs based on this update

        These will typically include updates to relationships on other nodes.
        """
        schema_branch = self._db.schema.get_schema_branch(name=self._branch.name)
        node_schema = schema_branch.get(name=primary_changelog.node_kind, duplicate=False)
        secondaries: list[NodeChangelog] = []

        for relationship in primary_changelog.relationships.values():
            if isinstance(relationship, RelationshipCardinalityOneChangelog):
                secondaries.extend(
                    self._parse_cardinality_one_relationship(
                        relationship=relationship,
                        node_schema=node_schema,
                        primary_changelog=primary_changelog,
                        schema_branch=schema_branch,
                    )
                )
            elif isinstance(relationship, RelationshipCardinalityManyChangelog):
                secondaries.extend(
                    self._parse_cardinality_many_relationship(
                        relationship=relationship,
                        node_schema=node_schema,
                        primary_changelog=primary_changelog,
                        schema_branch=schema_branch,
                    )
                )

        return secondaries

    def _parse_cardinality_one_relationship(
        self,
        relationship: RelationshipCardinalityOneChangelog,
        node_schema: MainSchemaTypes,
        primary_changelog: NodeChangelog,
        schema_branch: SchemaBranch,
    ) -> list[NodeChangelog]:
        secondaries: list[NodeChangelog] = []
        rel_schema = node_schema.get_relationship(name=relationship.name)

        if relationship.peer_status == DiffAction.ADDED:
            peer_schema = schema_branch.get(name=str(relationship.peer_kind), duplicate=False)
            secondaries.extend(
                self._process_added_peers(
                    peer_id=str(relationship.peer_id),
                    peer_kind=str(relationship.peer_kind),
                    peer_schema=peer_schema,
                    rel_schema=rel_schema,
                    primary_changelog=primary_changelog,
                )
            )

        elif relationship.peer_status == DiffAction.UPDATED:
            peer_schema = schema_branch.get(name=str(relationship.peer_kind), duplicate=False)
            secondaries.extend(
                self._process_added_peers(
                    peer_id=str(relationship.peer_id),
                    peer_kind=str(relationship.peer_kind),
                    peer_schema=peer_schema,
                    rel_schema=rel_schema,
                    primary_changelog=primary_changelog,
                )
            )
            secondaries.extend(
                self._process_removed_peers(
                    peer_schema=peer_schema,
                    peer_id=str(relationship.peer_id_previous),
                    peer_kind=str(relationship.peer_kind_previous),
                    rel_schema=rel_schema,
                    primary_changelog=primary_changelog,
                )
            )

        elif relationship.peer_status == DiffAction.REMOVED:
            peer_schema = schema_branch.get(name=str(relationship.peer_kind_previous), duplicate=False)

            secondaries.extend(
                self._process_removed_peers(
                    peer_id=str(relationship.peer_id_previous),
                    peer_kind=str(relationship.peer_kind_previous),
                    peer_schema=peer_schema,
                    rel_schema=rel_schema,
                    primary_changelog=primary_changelog,
                )
            )

        return secondaries

    def _parse_cardinality_many_relationship(
        self,
        relationship: RelationshipCardinalityManyChangelog,
        node_schema: MainSchemaTypes,
        primary_changelog: NodeChangelog,
        schema_branch: SchemaBranch,
    ) -> list[NodeChangelog]:
        secondaries: list[NodeChangelog] = []
        rel_schema = node_schema.get_relationship(name=relationship.name)

        for peer in relationship.peers:
            if peer.peer_status == DiffAction.ADDED:
                peer_schema = schema_branch.get(name=peer.peer_kind)
                secondaries.extend(
                    self._process_added_peers(
                        peer_id=peer.peer_id,
                        peer_kind=peer.peer_kind,
                        peer_schema=peer_schema,
                        rel_schema=rel_schema,
                        primary_changelog=primary_changelog,
                    )
                )

            elif peer.peer_status == DiffAction.REMOVED:
                peer_schema = schema_branch.get(name=peer.peer_kind)
                secondaries.extend(
                    self._process_removed_peers(
                        peer_id=peer.peer_id,
                        peer_kind=peer.peer_kind,
                        peer_schema=peer_schema,
                        rel_schema=rel_schema,
                        primary_changelog=primary_changelog,
                    )
                )

        return secondaries

    def _process_added_peers(
        self,
        peer_id: str,
        peer_kind: str,
        peer_schema: MainSchemaTypes,
        rel_schema: RelationshipSchema,
        primary_changelog: NodeChangelog,
    ) -> list[NodeChangelog]:
        secondaries: list[NodeChangelog] = []
        peer_relation = peer_schema.get_relationship_by_identifier(id=str(rel_schema.identifier), raise_on_error=False)
        if peer_relation:
            node_changelog = NodeChangelog(
                node_id=peer_id,
                node_kind=peer_kind,
                display_label="n/a",
            )
            if peer_relation.cardinality == RelationshipCardinality.ONE:
                node_changelog.relationships[peer_relation.name] = RelationshipCardinalityOneChangelog(
                    name=peer_relation.name,
                    peer_id=primary_changelog.node_id,
                    peer_kind=primary_changelog.node_kind,
                )
                secondaries.append(node_changelog)
            elif peer_relation.cardinality == RelationshipCardinality.MANY:
                node_changelog.relationships[peer_relation.name] = RelationshipCardinalityManyChangelog(
                    name=peer_relation.name,
                    peers=[
                        RelationshipPeerChangelog(
                            peer_id=primary_changelog.node_id,
                            peer_kind=primary_changelog.node_kind,
                            peer_status=DiffAction.ADDED,
                        )
                    ],
                )
                secondaries.append(node_changelog)

        return secondaries

    def _process_removed_peers(
        self,
        peer_id: str,
        peer_kind: str,
        peer_schema: MainSchemaTypes,
        rel_schema: RelationshipSchema,
        primary_changelog: NodeChangelog,
    ) -> list[NodeChangelog]:
        secondaries: list[NodeChangelog] = []
        peer_relation = peer_schema.get_relationship_by_identifier(id=str(rel_schema.identifier), raise_on_error=False)
        if peer_relation:
            node_changelog = NodeChangelog(
                node_id=peer_id,
                node_kind=peer_kind,
                display_label="n/a",
            )
            if peer_relation.cardinality == RelationshipCardinality.ONE:
                node_changelog.relationships[peer_relation.name] = RelationshipCardinalityOneChangelog(
                    name=peer_relation.name,
                    peer_id_previous=primary_changelog.node_id,
                    peer_kind_previous=primary_changelog.node_kind,
                )
                secondaries.append(node_changelog)
            elif peer_relation.cardinality == RelationshipCardinality.MANY:
                node_changelog.relationships[peer_relation.name] = RelationshipCardinalityManyChangelog(
                    name=peer_relation.name,
                    peers=[
                        RelationshipPeerChangelog(
                            peer_id=primary_changelog.node_id,
                            peer_kind=primary_changelog.node_kind,
                            peer_status=DiffAction.REMOVED,
                        )
                    ],
                )
                secondaries.append(node_changelog)

        return secondaries



from pydantic import BaseModel


class BranchCreateModel(BaseModel):
    name: str
    id: str | None = None
    description: str = ""
    origin_branch: str = "main"
    branched_from: str | None = None
    sync_with_git: bool = True
    is_isolated: bool = True



from infrahub.exceptions import FileOutOfRepositoryError, RepositoryFileNotFoundError
from infrahub.git.repository import get_initialized_repo
from infrahub.log import get_logger
from infrahub.message_bus import messages
from infrahub.message_bus.messages.git_file_get import (
    GitFileGetResponse,
    GitFileGetResponseData,
)
from infrahub.services import InfrahubServices

log = get_logger()


async def get(message: messages.GitFileGet, service: InfrahubServices) -> None:
    log.info("Collecting file from repository", repository=message.repository_name, file=message.file)

    repo = await get_initialized_repo(
        repository_id=message.repository_id,
        name=message.repository_name,
        service=service,
        repository_kind=message.repository_kind,
        commit=message.commit,
    )

    try:
        content = await repo.get_file(commit=message.commit, location=message.file)
    except (FileOutOfRepositoryError, RepositoryFileNotFoundError) as e:
        if message.reply_requested:
            response = GitFileGetResponse(data=GitFileGetResponseData(error_message=e.message, http_code=e.HTTP_CODE))
            await service.message_bus.reply_if_initiator_meta(message=response, initiator=message)
    else:
        if message.reply_requested:
            response = GitFileGetResponse(data=GitFileGetResponseData(content=content))
            await service.message_bus.reply_if_initiator_meta(message=response, initiator=message)



from typing import Any

from infrahub.core.constants import HashableModelState

deprecated_models: dict[str, Any] = {
    "generics": [
        {
            "name": "Source",
            "namespace": "Lineage",
            "display_labels": [],
            "attributes": [
                {"name": "name", "kind": "Text", "state": HashableModelState.ABSENT},
                {"name": "description", "kind": "Text", "state": HashableModelState.ABSENT},
            ],
        },
        {
            "name": "Owner",
            "namespace": "Lineage",
            "display_labels": [],
            "attributes": [
                {"name": "name", "kind": "Text", "state": HashableModelState.ABSENT},
                {"name": "description", "kind": "Text", "state": HashableModelState.ABSENT},
            ],
        },
    ]
}



from typing import Any

from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase

from ..model.path import TrackingId


class EnrichedDiffMergedTrackingIdQuery(Query):
    name = "enriched_diff_merge_tracking_id"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, tracking_ids: list[TrackingId], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.tracking_ids = tracking_ids

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {"tracking_ids": [t_id.serialize() for t_id in self.tracking_ids]}
        query = """
        MATCH (d_root:DiffRoot)
        WHERE d_root.tracking_id IN $tracking_ids
        SET d_root.is_merged = TRUE
        """
        self.add_to_query(query=query)



from infrahub.log import get_logger
from infrahub.message_bus import InfrahubMessage
from infrahub.message_bus.types import KVTTL
from infrahub.services import InfrahubServices

log = get_logger()


async def set_check_status(message: InfrahubMessage, conclusion: str, service: InfrahubServices) -> None:
    if message.meta.validator_execution_id and message.meta.check_execution_id:
        key = f"validator_execution_id:{message.meta.validator_execution_id}:check_execution_id:{message.meta.check_execution_id}"
        await service.cache.set(
            key=key,
            value=conclusion,
            expires=KVTTL.TWO_HOURS,
        )
        log.debug("setting check status in cache", key=key, conclusion=conclusion)



from pydantic import BaseModel, Field


class RequestGraphQLQueryGroupUpdate(BaseModel):
    """Sent to create or update a GraphQLQueryGroup associated with a given GraphQLQuery."""

    branch: str = Field(..., description="The branch to target")
    query_name: str = Field(..., description="The name of the GraphQLQuery that should be associated with the group")
    query_id: str = Field(..., description="The ID of the GraphQLQuery that should be associated with the group")
    related_node_ids: list[str] = Field(..., description="List of nodes related to the GraphQLQuery")
    subscribers: list[str] = Field(..., description="List of subscribers to add to the group")
    params: dict[str, str] = Field(..., description="Params sent with the query")



from ..models import EdgeToAdd, EdgeToDelete, PatchPlan, VertexToDelete
from .base import PatchQuery


class ConsolidateDuplicatedNodesPatchQuery(PatchQuery):
    """
    Find any groups of nodes with the same labels and properties, move all the edges to one of the duplicated nodes,
    then delete the other duplicated nodes
    """

    @property
    def name(self) -> str:
        return "consolidate-duplicated-nodes"

    async def plan(self) -> PatchPlan:
        query = """
//------------
// Find nodes with the same labels and UUID
//------------
MATCH (n:Node)
WITH n.uuid AS node_uuid, count(*) as num_nodes_with_uuid
WHERE num_nodes_with_uuid > 1
WITH DISTINCT node_uuid
MATCH (n:Node {uuid: node_uuid})
CALL {
    WITH n
    WITH labels(n) AS n_labels
    UNWIND n_labels AS n_label
    WITH n_label
    ORDER BY n_label ASC
    RETURN collect(n_label) AS sorted_labels
}
WITH n.uuid AS n_uuid, sorted_labels, collect(n) AS duplicate_nodes
WHERE size(duplicate_nodes) > 1
WITH n_uuid, head(duplicate_nodes) AS node_to_keep, tail(duplicate_nodes) AS nodes_to_delete
UNWIND nodes_to_delete AS node_to_delete
//------------
// Find the edges that we need to move to the selected node_to_keep
//------------
CALL {
    WITH node_to_keep, node_to_delete
    MATCH (node_to_delete)-[edge_to_delete]->(peer)
    RETURN {
        from_id: %(id_func_name)s(node_to_keep),
        to_id: %(id_func_name)s(peer),
        edge_type: type(edge_to_delete),
        after_props: properties(edge_to_delete)
    } AS edge_to_create
    UNION
    WITH node_to_keep, node_to_delete
    MATCH (node_to_delete)<-[edge_to_delete]-(peer)
    RETURN {
        from_id: %(id_func_name)s(peer),
        to_id: %(id_func_name)s(node_to_keep),
        edge_type: type(edge_to_delete),
        after_props: properties(edge_to_delete)
    } AS edge_to_create
}
WITH node_to_delete, collect(edge_to_create) AS edges_to_create
//------------
// Find the edges that we need to remove from the duplicated nodes
//------------
CALL {
    WITH node_to_delete
    MATCH (node_to_delete)-[e]->(peer)
    RETURN {
        db_id: %(id_func_name)s(e),
        from_id: %(id_func_name)s(node_to_delete),
        to_id: %(id_func_name)s(peer),
        edge_type: type(e),
        before_props: properties(e)
    } AS edge_to_delete
    UNION
    WITH node_to_delete
    MATCH (node_to_delete)<-[e]-(peer)
    RETURN {
        db_id: %(id_func_name)s(e),
        from_id: %(id_func_name)s(peer),
        to_id: %(id_func_name)s(node_to_delete),
        edge_type: type(e),
        before_props: properties(e)
    } AS edge_to_delete
}
WITH node_to_delete, edges_to_create, collect(edge_to_delete) AS edges_to_delete
RETURN
    {db_id: %(id_func_name)s(node_to_delete), labels: labels(node_to_delete), before_props: properties(node_to_delete)} AS vertex_to_delete,
    edges_to_create,
    edges_to_delete
        """ % {"id_func_name": self.db.get_id_function_name()}
        results = await self.db.execute_query(query=query)
        vertices_to_delete: list[VertexToDelete] = []
        edges_to_delete: list[EdgeToDelete] = []
        edges_to_add: list[EdgeToAdd] = []
        for result in results:
            serial_vertex_to_delete = result.get("vertex_to_delete")
            if serial_vertex_to_delete:
                vertex_to_delete = VertexToDelete(**serial_vertex_to_delete)
                vertices_to_delete.append(vertex_to_delete)
            for serial_edge_to_delete in result.get("edges_to_delete"):
                edge_to_delete = EdgeToDelete(**serial_edge_to_delete)
                edges_to_delete.append(edge_to_delete)
            for serial_edge_to_create in result.get("edges_to_create"):
                edges_to_add.append(EdgeToAdd(**serial_edge_to_create))
        return PatchPlan(
            name=self.name,
            vertices_to_delete=vertices_to_delete,
            edges_to_add=edges_to_add,
            edges_to_delete=edges_to_delete,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, Field, InputField, InputObjectType, Int, List, Mutation, String
from graphene.types.generic import GenericScalar
from typing_extensions import Self

from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.core.ipam.constants import PrefixMemberType
from infrahub.core.schema import NodeSchema
from infrahub.database import retry_db_transaction
from infrahub.exceptions import QueryValidationError, SchemaNotFoundError, ValidationError

from ..queries.resource_manager import PoolAllocatedNode
from .main import InfrahubMutationMixin, InfrahubMutationOptions

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.node.resource_manager.ip_address_pool import CoreIPAddressPool
    from infrahub.core.node.resource_manager.ip_prefix_pool import CoreIPPrefixPool
    from infrahub.database import InfrahubDatabase

    from ..initialization import GraphqlContext


class IPPrefixPoolGetResourceInput(InputObjectType):
    id = InputField(String(required=False), description="ID of the pool to allocate from")
    hfid = InputField(List(of_type=String, required=False), description="HFID of the pool to allocate from")
    identifier = InputField(String(required=False), description="Identifier for the allocated resource")
    prefix_length = InputField(Int(required=False), description="Size of the prefix to allocate")
    member_type = InputField(String(required=False), description="Type of members for the newly created prefix")
    prefix_type = InputField(String(required=False), description="Kind of prefix to allocate")
    data = InputField(GenericScalar(required=False), description="Additional data to pass to the newly created prefix")


class IPAddressPoolGetResourceInput(InputObjectType):
    id = InputField(String(required=False), description="ID of the pool to allocate from")
    hfid = InputField(List(of_type=String, required=False), description="HFID of the pool to allocate from")
    identifier = InputField(String(required=False), description="Identifier for the allocated resource")
    prefix_length = InputField(
        Int(required=False), description="Size of the prefix mask to allocate on the new IP address"
    )
    address_type = InputField(String(required=False), description="Kind of IP address to allocate")
    data = InputField(
        GenericScalar(required=False), description="Additional data to pass to the newly created IP address"
    )


class IPPrefixPoolGetResource(Mutation):
    class Arguments:
        data = IPPrefixPoolGetResourceInput(required=True)

    ok = Boolean()
    node = Field(PoolAllocatedNode)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: InputObjectType,
    ) -> Self:
        graphql_context: GraphqlContext = info.context

        member_type = data.get("member_type", None)
        allowed_member_types = [t.value for t in PrefixMemberType]
        if member_type and member_type not in allowed_member_types:
            raise QueryValidationError(f"Invalid member_type value, allowed values are {allowed_member_types}")

        obj: CoreIPPrefixPool = await registry.manager.find_object(  # type: ignore[assignment]
            db=graphql_context.db,
            kind=InfrahubKind.IPPREFIXPOOL,
            id=data.get("id"),
            hfid=data.get("hfid"),
            branch=graphql_context.branch,
        )
        resource = await obj.get_resource(
            db=graphql_context.db,
            branch=graphql_context.branch,
            identifier=data.get("identifier", None),
            prefixlen=data.get("prefix_length", None),
            member_type=member_type,
            prefix_type=data.get("prefix_type", None),
            data=data.get("data", None),
        )

        result = {
            "ok": True,
            "node": {
                "id": resource.id,
                "kind": resource.get_kind(),
                "identifier": data.get("identifier", None),
                "display_label": await resource.render_display_label(db=graphql_context.db),
                "branch": graphql_context.branch.name,
            },
        }

        return cls(**result)


class IPAddressPoolGetResource(Mutation):
    class Arguments:
        data = IPAddressPoolGetResourceInput(required=True)

    ok = Boolean()
    node = Field(PoolAllocatedNode)

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: dict[str, Any],
    ) -> Self:
        graphql_context: GraphqlContext = info.context

        obj: CoreIPAddressPool = await registry.manager.find_object(
            db=graphql_context.db,
            kind=InfrahubKind.IPADDRESSPOOL,
            id=data.get("id"),
            hfid=data.get("hfid"),
            branch=graphql_context.branch,
        )
        resource = await obj.get_resource(
            db=graphql_context.db,
            branch=graphql_context.branch,
            identifier=data.get("identifier"),
            prefixlen=data.get("prefix_length"),
            address_type=data.get("address_type"),
            data=data.get("data"),
        )

        result = {
            "ok": True,
            "node": {
                "id": resource.id,
                "kind": resource.get_kind(),
                "identifier": data.get("identifier"),
                "display_label": await resource.render_display_label(db=graphql_context.db),
                "branch": graphql_context.branch.name,
            },
        }

        return cls(**result)


class InfrahubNumberPoolMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema | None = None,
        _meta: InfrahubMutationOptions | None = None,
        **options: Any,
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")
        if not _meta:
            _meta = InfrahubMutationOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    @retry_db_transaction(name="resource_manager_create")
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,  # noqa: ARG003
    ) -> Any:
        try:
            pool_node = registry.schema.get(name=data["node"].value)
            if not pool_node.is_generic_schema and not pool_node.is_node_schema:
                raise ValidationError(input_value="The selected model is not a Node or a Generic")
        except SchemaNotFoundError as exc:
            exc.message = "The selected model does not exist"
            raise exc

        attributes = [attribute for attribute in pool_node.attributes if attribute.name == data["node_attribute"].value]
        if not attributes:
            raise ValidationError(input_value="The selected attribute doesn't exist in the selected model")

        attribute = attributes[0]
        if attribute.kind != "Number":
            raise ValidationError(input_value="The selected attribute is not of the kind Number")

        if data["start_range"].value > data["end_range"].value:
            raise ValidationError(input_value="start_range can't be larger than end_range")

        return await super().mutate_create(info=info, data=data, branch=branch)

    @classmethod
    @retry_db_transaction(name="resource_manager_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,  # noqa: ARG003
        node: Node | None = None,
    ) -> tuple[Node, Self]:
        if (data.get("node") and data.get("node").value) or (
            data.get("node_attribute") and data.get("node_attribute").value
        ):
            raise ValidationError(input_value="The fields 'node' or 'node_attribute' can't be changed.")
        graphql_context: GraphqlContext = info.context

        async with graphql_context.db.start_transaction() as dbt:
            number_pool, result = await super().mutate_update(
                info=info, data=data, branch=branch, database=dbt, node=node
            )
            if number_pool.start_range.value > number_pool.end_range.value:  # type: ignore[attr-defined]
                raise ValidationError(input_value="start_range can't be larger than end_range")

        return number_pool, result



from typing import Any

from infrahub import config

from ..app import InfrahubGraphQLApp
from ..auth.query_permission_checker.anonymous_checker import AnonymousGraphQLPermissionChecker
from ..auth.query_permission_checker.checker import GraphQLQueryPermissionChecker
from ..auth.query_permission_checker.default_branch_checker import DefaultBranchPermissionChecker
from ..auth.query_permission_checker.merge_operation_checker import MergeBranchPermissionChecker
from ..auth.query_permission_checker.object_permission_checker import (
    AccountManagerPermissionChecker,
    ObjectPermissionChecker,
    PermissionManagerPermissionChecker,
    RepositoryManagerPermissionChecker,
)
from ..auth.query_permission_checker.super_admin_checker import SuperAdminPermissionChecker


def get_anonymous_access_setting() -> bool:
    return config.SETTINGS.main.allow_anonymous_access


def build_graphql_query_permission_checker() -> GraphQLQueryPermissionChecker:
    return GraphQLQueryPermissionChecker(
        [
            AnonymousGraphQLPermissionChecker(get_anonymous_access_setting),
            # This checker never raises, it either terminates the checker chains (user is super admin) or go to the next one
            SuperAdminPermissionChecker(),
            DefaultBranchPermissionChecker(),
            MergeBranchPermissionChecker(),
            AccountManagerPermissionChecker(),
            RepositoryManagerPermissionChecker(),
            PermissionManagerPermissionChecker(),
            ObjectPermissionChecker(),
        ]
    )


def build_graphql_app(**kwargs: Any) -> InfrahubGraphQLApp:
    return InfrahubGraphQLApp(build_graphql_query_permission_checker(), **kwargs)



import uuid

from pydantic import Field

from infrahub.context import InfrahubContext
from infrahub.core.constants import CheckType
from infrahub.message_bus import InfrahubMessage


class RequestProposedChangePipeline(InfrahubMessage):
    """Sent request the start of a pipeline connected to a proposed change."""

    proposed_change: str = Field(..., description="The unique ID of the proposed change")
    source_branch: str = Field(..., description="The source branch of the proposed change")
    source_branch_sync_with_git: bool = Field(..., description="Indicates if the source branch should sync with git")
    destination_branch: str = Field(..., description="The destination branch of the proposed change")
    check_type: CheckType = Field(
        default=CheckType.ALL, description="Can be used to restrict the pipeline to a specific type of job"
    )
    context: InfrahubContext = Field(..., description="The context of the task")
    pipeline_id: uuid.UUID = Field(
        default_factory=uuid.uuid4, description="The unique ID of the execution of this pipeline"
    )



from __future__ import annotations

import os
from typing import TYPE_CHECKING

from .repository import GitRepoManager

if TYPE_CHECKING:
    from . import InfrahubClient
    from .node import InfrahubNode
    from .store import NodeStore


class InfrahubOperation:
    def __init__(
        self,
        client: InfrahubClient,
        infrahub_node: type[InfrahubNode],
        convert_query_response: bool,
        branch: str,
        root_directory: str,
    ):
        self.branch = branch
        self.convert_query_response = convert_query_response
        self.root_directory = root_directory or os.getcwd()
        self.infrahub_node = infrahub_node
        self._nodes: list[InfrahubNode] = []
        self._related_nodes: list[InfrahubNode] = []
        self._init_client = client.clone(branch=self.branch_name)
        self.git: GitRepoManager | None = None

    @property
    def branch_name(self) -> str:
        """Return the name of the current git branch."""

        if self.branch:
            return self.branch

        if not hasattr(self, "git") or not self.git:
            self.git = GitRepoManager(self.root_directory)

        self.branch = str(self.git.active_branch)

        return self.branch

    @property
    def store(self) -> NodeStore:
        """The store will be populated with nodes based on the query during the collection of data if activated"""
        return self._init_client.store

    @property
    def nodes(self) -> list[InfrahubNode]:
        """Returns nodes collected and parsed during the data collection process if this feature is enabled"""
        return self._nodes

    @property
    def related_nodes(self) -> list[InfrahubNode]:
        """Returns nodes collected and parsed during the data collection process if this feature is enabled"""
        return self._related_nodes

    async def process_nodes(self, data: dict) -> None:
        if not self.convert_query_response:
            return

        await self._init_client.schema.all(branch=self.branch_name)

        for kind in data:
            if kind in self._init_client.schema.cache[self.branch_name].nodes.keys():
                for result in data[kind].get("edges", []):
                    node = await self.infrahub_node.from_graphql(
                        client=self._init_client, branch=self.branch_name, data=result
                    )
                    self._nodes.append(node)
                    await node._process_relationships(
                        node_data=result, branch=self.branch_name, related_nodes=self._related_nodes
                    )

        for node in self._nodes + self._related_nodes:
            if node.id:
                self._init_client.store.set(node=node)



from __future__ import annotations

from datetime import datetime
from enum import Enum

from pydantic import BaseModel, Field


class TaskState(str, Enum):
    SCHEDULED = "SCHEDULED"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"
    CRASHED = "CRASHED"
    PAUSED = "PAUSED"
    CANCELLING = "CANCELLING"


class TaskLog(BaseModel):
    message: str
    severity: str
    timestamp: datetime


class TaskRelatedNode(BaseModel):
    id: str
    kind: str


class Task(BaseModel):
    id: str
    title: str
    state: TaskState
    progress: float | None = None
    workflow: str | None = None
    branch: str | None = None
    # start_time: datetime # Is it still required
    created_at: datetime
    updated_at: datetime
    parameters: dict | None = None
    tags: list[str] | None = None
    related_nodes: list[TaskRelatedNode] = Field(default_factory=list)
    logs: list[TaskLog] = Field(default_factory=list)

    @classmethod
    def from_graphql(cls, data: dict) -> Task:
        related_nodes: list[TaskRelatedNode] = []
        logs: list[TaskLog] = []

        if data.get("related_nodes"):
            related_nodes = [TaskRelatedNode(**item) for item in data["related_nodes"]]
            del data["related_nodes"]

        if data.get("logs"):
            logs = [TaskLog(**item["node"]) for item in data["logs"]["edges"]]
            del data["logs"]

        return cls(**data, related_nodes=related_nodes, logs=logs)


class TaskFilter(BaseModel):
    ids: list[str] | None = None
    q: str | None = None
    branch: str | None = None
    state: list[TaskState] | None = None
    workflow: list[str] | None = None
    limit: int | None = None
    offset: int | None = None
    related_node__ids: list[str] | None = None

    def to_dict(self) -> dict:
        return self.model_dump(exclude_none=True)



import json
from dataclasses import asdict
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

from .constants import PatchPlanFilename
from .models import EdgeToAdd, EdgeToDelete, EdgeToUpdate, PatchPlan, VertexToAdd, VertexToDelete, VertexToUpdate


class PatchPlanWriter:
    def write(self, patches_directory: Path, patch_plan: PatchPlan) -> Path:
        timestamp_str = datetime.now(tz=timezone.utc).strftime("%Y%m%d-%H%M%S")
        patch_name = f"patch-{patch_plan.name}-{timestamp_str}"
        patch_plan_directory = patches_directory / Path(patch_name)
        if not patch_plan_directory.exists():
            patch_plan_directory.mkdir(parents=True)
        if patch_plan.vertices_to_add:
            self._write_vertices_to_add(
                patch_plan_directory=patch_plan_directory, vertices_to_add=patch_plan.vertices_to_add
            )
        if patch_plan.vertices_to_delete:
            self._write_vertices_to_delete(
                patch_plan_directory=patch_plan_directory, vertices_to_delete=patch_plan.vertices_to_delete
            )
        if patch_plan.vertices_to_update:
            self._write_vertices_to_update(
                patch_plan_directory=patch_plan_directory, vertices_to_update=patch_plan.vertices_to_update
            )
        if patch_plan.edges_to_add:
            self._write_edges_to_add(patch_plan_directory=patch_plan_directory, edges_to_add=patch_plan.edges_to_add)
        if patch_plan.edges_to_delete:
            self._write_edges_to_delete(
                patch_plan_directory=patch_plan_directory, edges_to_delete=patch_plan.edges_to_delete
            )
        if patch_plan.edges_to_update:
            self._write_edges_to_update(
                patch_plan_directory=patch_plan_directory, edges_to_update=patch_plan.edges_to_update
            )

        return patch_plan_directory

    def write_added_db_id_map(self, patch_plan_directory: Path, db_id_map: dict[str, str]) -> None:
        file = patch_plan_directory / Path(PatchPlanFilename.ADDED_DB_IDS.value)
        file.touch(exist_ok=True)
        with file.open(mode="w") as f:
            f.write(json.dumps(db_id_map) + "\n")

    def write_deleted_db_ids(self, patch_plan_directory: Path, deleted_ids: set[str]) -> None:
        file = patch_plan_directory / Path(PatchPlanFilename.DELETED_DB_IDS.value)
        file.touch(exist_ok=True)
        with file.open(mode="w") as f:
            f.write(json.dumps(list(deleted_ids)) + "\n")

    def write_reverted_deleted_db_id_map(self, patch_plan_directory: Path, db_id_map: dict[str, str]) -> None:
        file = patch_plan_directory / Path(PatchPlanFilename.REVERTED_DELETED_DB_IDS.value)
        file.touch(exist_ok=True)
        with file.open(mode="w") as f:
            f.write(json.dumps(db_id_map) + "\n")

    def _dataclass_to_json_line(self, dataclass_instance: Any) -> str:
        return json.dumps(asdict(dataclass_instance)) + "\n"

    def _write_to_file(self, file_path: Path, objects: list[Any]) -> None:
        file_path.touch(exist_ok=True)
        with file_path.open(mode="w") as f:
            for obj in objects:
                f.write(self._dataclass_to_json_line(obj))

    def _write_vertices_to_add(self, patch_plan_directory: Path, vertices_to_add: list[VertexToAdd]) -> None:
        file = patch_plan_directory / Path(PatchPlanFilename.VERTICES_TO_ADD.value)
        self._write_to_file(file_path=file, objects=vertices_to_add)

    def _write_vertices_to_delete(self, patch_plan_directory: Path, vertices_to_delete: list[VertexToDelete]) -> None:
        file = patch_plan_directory / Path(PatchPlanFilename.VERTICES_TO_DELETE.value)
        self._write_to_file(file_path=file, objects=vertices_to_delete)

    def _write_vertices_to_update(self, patch_plan_directory: Path, vertices_to_update: list[VertexToUpdate]) -> None:
        file = patch_plan_directory / Path(PatchPlanFilename.VERTICES_TO_UPDATE.value)
        self._write_to_file(file_path=file, objects=vertices_to_update)

    def _write_edges_to_add(self, patch_plan_directory: Path, edges_to_add: list[EdgeToAdd]) -> None:
        file = patch_plan_directory / Path(PatchPlanFilename.EDGES_TO_ADD.value)
        self._write_to_file(file_path=file, objects=edges_to_add)

    def _write_edges_to_delete(self, patch_plan_directory: Path, edges_to_delete: list[EdgeToDelete]) -> None:
        file = patch_plan_directory / Path(PatchPlanFilename.EDGES_TO_DELETE.value)
        self._write_to_file(file_path=file, objects=edges_to_delete)

    def _write_edges_to_update(self, patch_plan_directory: Path, edges_to_update: list[EdgeToUpdate]) -> None:
        file = patch_plan_directory / Path(PatchPlanFilename.EDGES_TO_UPDATE.value)
        self._write_to_file(file_path=file, objects=edges_to_update)



from __future__ import annotations

from typing import TYPE_CHECKING

from graphene.types.definitions import GrapheneInterfaceType, GrapheneObjectType
from graphql import (
    GraphQLList,
    GraphQLNonNull,
    GraphQLSchema,
    GraphQLUnionType,
)

from infrahub.exceptions import GraphQLQueryError

if TYPE_CHECKING:
    from graphql.execution import ExecutionResult


def extract_data(query_name: str, result: ExecutionResult) -> dict:
    if result.errors:
        errors = []
        for error in result.errors:
            error_locations = error.locations or []
            errors.append(
                {
                    "message": f"GraphQLQuery {query_name}: {error.message}",
                    "path": error.path,
                    "locations": [{"line": location.line, "column": location.column} for location in error_locations],
                }
            )

        raise GraphQLQueryError(errors=errors)

    return result.data or {}


def find_types_implementing_interface(
    interface: GrapheneInterfaceType, root_schema: GraphQLSchema
) -> list[GrapheneObjectType]:
    results = []
    for value in root_schema.type_map.values():
        if not hasattr(value, "interfaces"):
            continue

        for item in value.interfaces:
            if item.name == interface.name:
                results.append(value)

    return results


async def extract_schema_models(
    fields: dict, schema: GrapheneObjectType | GraphQLUnionType, root_schema: GraphQLSchema
) -> set[str]:
    response = set()

    if isinstance(schema, GraphQLUnionType):
        return {t.name for t in schema.types}

    for field_name, value in fields.items():
        if field_name not in schema.fields:
            continue

        gql_type = schema.fields[field_name].type
        if isinstance(schema.fields[field_name].type, GraphQLNonNull):
            gql_type = schema.fields[field_name].type.of_type

        if isinstance(gql_type, GrapheneObjectType):
            object_type = gql_type
        elif isinstance(gql_type, GraphQLList):
            if isinstance(gql_type.of_type, GraphQLNonNull):
                object_type = gql_type.of_type.of_type
            else:
                object_type = gql_type.of_type
        elif isinstance(gql_type, GrapheneInterfaceType):
            object_type = gql_type
            sub_types = find_types_implementing_interface(interface=object_type, root_schema=root_schema)
            for sub_type in sub_types:
                response.add(sub_type.name)
                response.update(await extract_schema_models(fields=value, schema=sub_type, root_schema=root_schema))
        else:
            continue

        # Ensure that Attribute types are not reported by this function
        if isinstance(object_type, GrapheneObjectType) and object_type.interfaces:
            inherit_from = [intf.name for intf in object_type.interfaces]
            if "AttributeInterface" in inherit_from:
                continue

        if isinstance(object_type, GraphQLNonNull):
            raise ValueError("object_type shouldn't be a of type GraphQLNonNull")

        response.add(object_type.name)

        if isinstance(value, dict):
            response.update(await extract_schema_models(fields=value, schema=object_type, root_schema=root_schema))
        elif isinstance(value, str) and value in schema.fields:
            if isinstance(schema.fields[value].type, GrapheneObjectType):
                response.add(schema.fields[value].type.name)
            elif isinstance(schema.fields[value].type, GraphQLList):
                if isinstance(schema.fields[value].type.of_type, GraphQLNonNull):
                    response.add(schema.fields[value].type.of_type.of_type.name)
                else:
                    response.add(schema.fields[value].type.of_type.name)

    return response



import io
import tempfile
from typing import Any, BinaryIO

import botocore.exceptions
import fastapi_storages
from typing_extensions import Self

from infrahub.config import StorageSettings
from infrahub.exceptions import NodeNotFoundError


class InfrahubS3ObjectStorage(fastapi_storages.S3Storage):
    def __init__(self, **kwargs: Any) -> None:
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
        super().__init__()

    def open(self, name: str) -> BinaryIO:
        f = io.BytesIO()
        self._bucket.download_fileobj(name, f)
        f.flush()
        f.seek(0)
        return f  # type: ignore


fastapi_storages.InfrahubS3ObjectStorage = InfrahubS3ObjectStorage


class InfrahubObjectStorage:
    _settings: StorageSettings
    _storage: fastapi_storages.base.BaseStorage

    def __init__(self, settings: StorageSettings) -> None:
        self._settings = settings

        driver = getattr(fastapi_storages, self._settings.driver.name)

        driver_settings = getattr(self._settings, self._settings.driver.value.lower())
        self._storage = driver(**driver_settings.model_dump(by_alias=True))

    @classmethod
    async def init(cls, settings: StorageSettings) -> Self:
        return cls(settings)

    def store(self, identifier: str, content: bytes) -> None:
        with tempfile.NamedTemporaryFile() as f:
            f.write(content)
            self._storage.write(f, identifier)

    def retrieve(self, identifier: str) -> str:
        try:
            with self._storage.open(identifier) as f:
                return f.read().decode()
        except (FileNotFoundError, botocore.exceptions.ClientError) as err:
            raise NodeNotFoundError(node_type="StorageObject", identifier=identifier) from err



from __future__ import annotations

from typing import TYPE_CHECKING

from ..interface import ConstraintCheckerInterface
from ..query import NodeNotPresentValidatorQuery

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.path import GroupedDataPaths
    from infrahub.database import InfrahubDatabase

    from ..model import SchemaConstraintValidatorRequest


class NodeAttributeAddChecker(ConstraintCheckerInterface):
    query_classes = [NodeNotPresentValidatorQuery]

    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None):
        self.db = db
        self.branch = branch

    @property
    def name(self) -> str:
        return "node.attribute.add"

    def supports(self, request: SchemaConstraintValidatorRequest) -> bool:
        return request.constraint_name == self.name

    async def check(self, request: SchemaConstraintValidatorRequest) -> list[GroupedDataPaths]:
        grouped_data_paths_list: list[GroupedDataPaths] = []
        if not request.schema_path.field_name:
            raise ValueError("field_name is not defined")
        attribute_schema = request.node_schema.get_attribute(name=request.schema_path.field_name)
        if attribute_schema.optional is True or attribute_schema.default_value is not None:
            return grouped_data_paths_list

        for query_class in self.query_classes:
            # TODO add exception handling
            query = await query_class.init(
                db=self.db, branch=self.branch, node_schema=request.node_schema, schema_path=request.schema_path
            )
            await query.execute(db=self.db)
            grouped_data_paths_list.append(await query.get_paths())
        return grouped_data_paths_list



from __future__ import annotations

from collections.abc import Mapping
from pathlib import Path

import jinja2

from .. import protocols as sdk_protocols
from ..schema import (
    AttributeSchemaAPI,
    GenericSchema,
    GenericSchemaAPI,
    MainSchemaTypesAll,
    NodeSchema,
    NodeSchemaAPI,
    ProfileSchemaAPI,
    RelationshipSchemaAPI,
    TemplateSchemaAPI,
)
from .constants import ATTRIBUTE_KIND_MAP, CORE_BASE_CLASS_TO_SYNCIFY, TEMPLATE_FILE_NAME


def load_template() -> str:
    path = Path(__file__).parent / TEMPLATE_FILE_NAME
    return path.read_text()


def move_to_end_of_list(lst: list, item: str) -> list:
    """Move an item to the end of a list if it exists in the list"""
    if item in lst:
        lst.remove(item)
        lst.append(item)
    return lst


class CodeGenerator:
    def __init__(self, schema: dict[str, MainSchemaTypesAll]):
        self.generics: dict[str, GenericSchemaAPI | GenericSchema] = {}
        self.nodes: dict[str, NodeSchemaAPI | NodeSchema] = {}
        self.profiles: dict[str, ProfileSchemaAPI] = {}
        self.templates: dict[str, TemplateSchemaAPI] = {}

        for name, schema_type in schema.items():
            if isinstance(schema_type, (GenericSchemaAPI, GenericSchema)):
                self.generics[name] = schema_type
            if isinstance(schema_type, (NodeSchemaAPI, NodeSchema)):
                self.nodes[name] = schema_type
            if isinstance(schema_type, ProfileSchemaAPI):
                self.profiles[name] = schema_type
            if isinstance(schema_type, TemplateSchemaAPI):
                self.templates[name] = schema_type

        self.base_protocols = [
            e
            for e in dir(sdk_protocols)
            if not e.startswith("__")
            and not e.endswith("__")
            and e
            not in ("TYPE_CHECKING", "CoreNode", "Optional", "Protocol", "Union", "annotations", "runtime_checkable")
        ]

        self.sorted_generics = self._sort_and_filter_models(self.generics, filters=["CoreNode"] + self.base_protocols)
        self.sorted_nodes = self._sort_and_filter_models(self.nodes, filters=["CoreNode"] + self.base_protocols)
        self.sorted_profiles = self._sort_and_filter_models(
            self.profiles, filters=["CoreProfile"] + self.base_protocols
        )
        self.sorted_templates = self._sort_and_filter_models(
            self.templates, filters=["CoreObjectTemplate"] + self.base_protocols
        )

    def render(self, sync: bool = True) -> str:
        jinja2_env = jinja2.Environment(loader=jinja2.BaseLoader(), trim_blocks=True, lstrip_blocks=True)
        jinja2_env.filters["render_attribute"] = self._jinja2_filter_render_attribute
        jinja2_env.filters["render_relationship"] = self._jinja2_filter_render_relationship
        jinja2_env.filters["syncify"] = self._jinja2_filter_syncify

        template = jinja2_env.from_string(load_template())
        return template.render(
            generics=self.sorted_generics,
            nodes=self.sorted_nodes,
            profiles=self.sorted_profiles,
            templates=self.sorted_templates,
            base_protocols=self.base_protocols,
            core_node_name="CoreNodeSync" if sync else "CoreNode",
            sync=sync,
        )

    @staticmethod
    def _jinja2_filter_syncify(value: str | list, sync: bool = False) -> str | list:
        """Filter to help with the convertion to sync

        If a string is provided, append Sync to the end of the string
        If a list is provided, search for CoreNode and replace it with CoreNodeSync
        """
        if isinstance(value, list):
            # Order the list based on the CORE_BASE_CLASS_TO_SYNCIFY list to ensure the base classes are always last
            for item in CORE_BASE_CLASS_TO_SYNCIFY:
                value = move_to_end_of_list(value, item)

        if not sync:
            return value

        if isinstance(value, str):
            return f"{value}Sync"

        if isinstance(value, list):
            return [f"{item}Sync" if item in CORE_BASE_CLASS_TO_SYNCIFY else item for item in value]

        return value

    @staticmethod
    def _jinja2_filter_render_attribute(value: AttributeSchemaAPI) -> str:
        attribute_kind: str = ATTRIBUTE_KIND_MAP[value.kind]

        if value.optional:
            attribute_kind += "Optional"

        return f"{value.name}: {attribute_kind}"

    @staticmethod
    def _jinja2_filter_render_relationship(value: RelationshipSchemaAPI, sync: bool = False) -> str:
        name = value.name
        cardinality = value.cardinality

        type_ = "RelatedNode"
        if cardinality == "many":
            type_ = "RelationshipManager"

        if sync:
            type_ += "Sync"

        return f"{name}: {type_}"

    @staticmethod
    def _sort_and_filter_models(
        models: Mapping[str, MainSchemaTypesAll], filters: list[str] | None = None
    ) -> list[MainSchemaTypesAll]:
        if filters is None:
            filters = ["CoreNode"]

        filtered: list[MainSchemaTypesAll] = []
        for name, model in models.items():
            if name in filters:
                continue
            filtered.append(model)

        return sorted(filtered, key=lambda k: k.name)



from __future__ import annotations

from typing import Any

from graphene import ObjectType
from graphene.types.objecttype import ObjectTypeOptions

from infrahub.core.schema import GenericSchema, MainSchemaTypes, NodeSchema, ProfileSchema, TemplateSchema


class InfrahubObjectOptions(ObjectTypeOptions):
    schema: MainSchemaTypes


class InfrahubObject(ObjectType):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: MainSchemaTypes | None = None,
        interfaces: tuple = (),
        _meta: InfrahubObjectOptions | None = None,
        **options: Any,
    ) -> None:
        if not isinstance(schema, NodeSchema | GenericSchema | ProfileSchema | TemplateSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubObjectOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, interfaces=interfaces, **options)



from dataclasses import dataclass

from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import RelationshipCardinality
from infrahub.core.query.node import NodeListGetInfoQuery
from infrahub.core.schema import MainSchemaTypes
from infrahub.core.schema.generic_schema import GenericSchema
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import ValidationError

from ..model import RelationshipManager
from .interface import RelationshipManagerConstraintInterface


@dataclass
class NodeToValidate:
    uuid: str
    cardinality: RelationshipCardinality
    min_count: int | None = None
    max_count: int | None = None


class RelationshipPeerKindConstraint(RelationshipManagerConstraintInterface):
    def __init__(self, db: InfrahubDatabase, branch: Branch | None = None):
        self.db = db
        self.branch = branch

    async def check(self, relm: RelationshipManager, node_schema: MainSchemaTypes) -> None:  # noqa: ARG002
        branch = await registry.get_branch(db=self.db) if not self.branch else self.branch
        peer_schema = registry.schema.get(name=relm.schema.peer, branch=branch, duplicate=False)
        if isinstance(peer_schema, GenericSchema):
            allowed_kinds = peer_schema.used_by
        else:
            allowed_kinds = [peer_schema.kind]
        relationships = await relm.get_relationships(db=self.db, force_refresh=False)
        if not relationships:
            return

        peer_ids = [r.peer_id for r in relationships]
        peers_query = await NodeListGetInfoQuery.init(db=self.db, branch=branch, ids=peer_ids)
        await peers_query.execute(db=self.db)

        errors: list[ValidationError] = []
        async for peer_node in peers_query.get_nodes(db=self.db, duplicate=False):
            if not peer_node.schema:
                raise ValueError(f"Cannot identify schema for node {peer_node.node_uuid}")
            if peer_node.schema.kind not in allowed_kinds:
                errors.append(
                    ValidationError(
                        {
                            relm.name: (
                                f"{peer_node.schema.kind} - {peer_node.node_uuid} cannot be added to relationship, "
                                f"must be of type: {allowed_kinds}"
                            )
                        }
                    )
                )

        if not errors:
            return

        raise ValidationError(errors)



from dataclasses import dataclass
from typing import Any

from aiodataloader import DataLoader

from infrahub.auth import AccountSession
from infrahub.core.branch.models import Branch
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase

from .shared import to_frozen_set


@dataclass
class GetManyParams:
    branch: Branch | str
    fields: dict | None = None
    at: Timestamp | str | None = None
    include_source: bool = False
    include_owner: bool = False
    prefetch_relationships: bool = False
    account: AccountSession | None = None
    branch_agnostic: bool = False

    def __hash__(self) -> int:
        frozen_fields: frozenset | None = None
        if self.fields:
            frozen_fields = to_frozen_set(self.fields)
        timestamp = Timestamp(self.at)
        branch = self.branch.name if isinstance(self.branch, Branch) else self.branch
        account_id = self.account.account_id if isinstance(self.account, AccountSession) else None
        hash_str = "|".join(
            [
                str(hash(frozen_fields)),
                timestamp.to_string(),
                branch,
                str(self.include_source),
                str(self.include_owner),
                str(self.prefetch_relationships),
                str(account_id),
                str(self.branch_agnostic),
            ]
        )
        return hash(hash_str)


class NodeDataLoader(DataLoader[str, Node | None]):
    def __init__(self, db: InfrahubDatabase, query_params: GetManyParams, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.query_params = query_params
        self.db = db

    async def batch_load_fn(self, keys: list[Any]) -> list[Node | None]:
        async with self.db.start_session() as db:
            nodes_by_id = await NodeManager.get_many(
                db=db,
                ids=keys,
                fields=self.query_params.fields,
                at=self.query_params.at,
                branch=self.query_params.branch,
                include_source=self.query_params.include_source,
                include_owner=self.query_params.include_owner,
                prefetch_relationships=self.query_params.prefetch_relationships,
                account=self.query_params.account,
                branch_agnostic=self.query_params.branch_agnostic,
            )
        results = []
        for node_id in keys:
            results.append(nodes_by_id.get(node_id, None))
        return results



from infrahub.context import InfrahubContext
from infrahub.core.branch import Branch
from infrahub.core.changelog.models import RelationshipChangelogGetter
from infrahub.core.constants import MutationAction
from infrahub.core.node import Node
from infrahub.database import InfrahubDatabase
from infrahub.events.node_action import NodeDeletedEvent, NodeMutatedEvent, NodeUpdatedEvent, get_node_event
from infrahub.groups.parsers import GroupNodeMutationParser
from infrahub.worker import WORKER_IDENTITY

from .models import EventMeta, InfrahubEvent


async def generate_node_mutation_events(
    node: Node,
    deleted_nodes: list[Node],
    db: InfrahubDatabase,
    branch: Branch,
    context: InfrahubContext,
    request_id: str,
    action: MutationAction,
) -> list[InfrahubEvent]:
    meta = EventMeta(
        account_id=context.account.account_id,
        initiator_id=WORKER_IDENTITY,
        request_id=request_id,
        branch=branch,
        context=context,
    )
    node_event_class = get_node_event(action)
    main_event = node_event_class(
        kind=node.get_kind(),
        node_id=node.id,
        changelog=node.node_changelog,
        fields=node.node_changelog.updated_fields,
        meta=meta,
    )
    relationship_changelogs = RelationshipChangelogGetter(db=db, branch=branch)
    node_changelogs = await relationship_changelogs.get_changelogs(primary_changelog=node.node_changelog)

    events: list[NodeMutatedEvent] = [main_event]

    deleted_changelogs = [deleted_node.node_changelog for deleted_node in deleted_nodes if deleted_node.id != node.id]
    deleted_ids = {deleted_node.node_id for deleted_node in deleted_changelogs}

    for node_changelog in deleted_changelogs:
        meta = EventMeta.from_parent(parent=main_event)
        delete_event = NodeDeletedEvent(
            kind=node_changelog.node_kind,
            node_id=node_changelog.node_id,
            changelog=node_changelog,
            fields=node_changelog.updated_fields,
            meta=meta,
        )
        events.append(delete_event)

    for node_changelog in node_changelogs:
        if node_changelog.node_id not in deleted_ids:
            meta = EventMeta.from_parent(parent=main_event)
            update_event = NodeUpdatedEvent(
                kind=node_changelog.node_kind,
                node_id=node_changelog.node_id,
                changelog=node_changelog,
                fields=node_changelog.updated_fields,
                meta=meta,
            )
            events.append(update_event)

    group_parser = GroupNodeMutationParser(db=db, branch=branch)
    group_events = await group_parser.group_events_from_node_actions(events=events)
    return events + group_events



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

from infrahub.core.branch import Branch
from infrahub.core.constants import GLOBAL_BRANCH_NAME, BranchSupportType, InfrahubKind
from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.query import Query, QueryType

from ..query.attribute_rename import AttributeInfo, AttributeRenameQuery
from ..query.delete_element_in_schema import DeleteElementInSchemaQuery
from ..query.node_duplicate import NodeDuplicateQuery, SchemaNodeInfo
from ..query.relationship_duplicate import RelationshipDuplicateQuery, SchemaRelationshipInfo
from ..query.schema_attribute_update import SchemaAttributeUpdateQuery
from ..shared import GraphMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


global_branch = Branch(
    name=GLOBAL_BRANCH_NAME,
    status="OPEN",
    description="Global Branch",
    hierarchy_level=1,
    is_global=True,
    sync_with_git=False,
)

default_branch = Branch(
    name="main",
    status="OPEN",
    description="Default Branch",
    hierarchy_level=1,
    is_global=False,
    is_default=True,
    sync_with_git=False,
)


class Migration012RenameTypeAttributeData(AttributeRenameQuery):
    name = "migration_012_rename_attr_type"
    type = QueryType.WRITE

    def __init__(self, **kwargs: Any):
        new_attr = AttributeInfo(
            name="account_type",
            node_kind=InfrahubKind.ACCOUNT,
            branch_support=BranchSupportType.AGNOSTIC.value,
        )
        previous_attr = AttributeInfo(
            name="type",
            node_kind=InfrahubKind.ACCOUNT,
            branch_support=BranchSupportType.AGNOSTIC.value,
        )

        kwargs.pop("branch", None)

        super().__init__(new_attr=new_attr, previous_attr=previous_attr, branch=global_branch, **kwargs)

    def render_match(self) -> str:
        query = """
        // Find all the active nodes
        CALL {
            MATCH (node:%(node_kind)s)
            WHERE exists((node)-[:HAS_ATTRIBUTE]-(:Attribute { name: $prev_attr.name }))
                AND NOT exists((node)-[:HAS_ATTRIBUTE]-(:Attribute { name: $new_attr.name }))
            RETURN node
            UNION
            MATCH (node:Profile%(node_kind)s)
            WHERE exists((node)-[:HAS_ATTRIBUTE]-(:Attribute { name: $prev_attr.name }))
                AND NOT exists((node)-[:HAS_ATTRIBUTE]-(:Attribute { name: $new_attr.name }))
            RETURN node
        }
        WITH node
        """ % {"node_kind": self.previous_attr.node_kind}

        return query


class Migration012AddLabelData(NodeDuplicateQuery):
    name = "migration_012_add_labels"
    type = QueryType.WRITE

    def __init__(self, **kwargs: Any):
        new_node = SchemaNodeInfo(
            name="Account",
            namespace="Core",
            branch_support=BranchSupportType.AGNOSTIC.value,
            labels=[
                InfrahubKind.ACCOUNT,
                InfrahubKind.GENERICACCOUNT,
                InfrahubKind.LINEAGEOWNER,
                InfrahubKind.LINEAGESOURCE,
            ],
            kind=InfrahubKind.ACCOUNT,
        )
        previous_node = SchemaNodeInfo(
            name="Account",
            namespace="Core",
            branch_support=BranchSupportType.AGNOSTIC.value,
            labels=[InfrahubKind.ACCOUNT, InfrahubKind.LINEAGEOWNER, InfrahubKind.LINEAGESOURCE],
            kind=InfrahubKind.ACCOUNT,
        )

        branch = Branch(
            name=GLOBAL_BRANCH_NAME,
            status="OPEN",
            description="Global Branch",
            hierarchy_level=1,
            is_global=True,
            sync_with_git=False,
        )

        kwargs.pop("branch", None)

        super().__init__(new_node=new_node, previous_node=previous_node, branch=branch, **kwargs)

    def render_match(self) -> str:
        query = f"""
        // Find all the active nodes
        MATCH (node:{self.previous_node.kind})
        WHERE NOT "CoreGenericAccount" IN LABELS(node)
        """

        return query


class Migration012RenameTypeAttributeSchema(SchemaAttributeUpdateQuery):
    name = "migration_012_rename_type_attr_schema"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, **kwargs: Any):
        super().__init__(
            attribute_name="name",
            node_name="Account",
            node_namespace="Core",
            new_value="account_type",
            previous_value="type",
            **kwargs,
        )

    def render_match(self) -> str:
        return self._render_match_schema_attribute()


class Migration012RenameRelationshipAccountTokenData(RelationshipDuplicateQuery):
    name = "migration_012_rename_rel_account_token_data"
    type = QueryType.WRITE

    def __init__(self, **kwargs: Any):
        new_rel = SchemaRelationshipInfo(
            name="account__token",
            branch_support=BranchSupportType.AGNOSTIC.value,
            src_peer=InfrahubKind.ACCOUNT,
            dst_peer=InfrahubKind.ACCOUNTTOKEN,
        )
        previous_rel = SchemaRelationshipInfo(
            name="coreaccount__internalaccounttoken",
            branch_support=BranchSupportType.AGNOSTIC.value,
            src_peer=InfrahubKind.ACCOUNT,
            dst_peer=InfrahubKind.ACCOUNTTOKEN,
        )

        kwargs.pop("branch", None)

        super().__init__(new_rel=new_rel, previous_rel=previous_rel, branch=global_branch, **kwargs)


class Migration012RenameRelationshipRefreshTokenData(RelationshipDuplicateQuery):
    name = "migration_012_rename_rel_refresh_token_data"
    type = QueryType.WRITE

    def __init__(self, **kwargs: Any):
        new_rel = SchemaRelationshipInfo(
            name="account__refreshtoken",
            branch_support=BranchSupportType.AGNOSTIC.value,
            src_peer=InfrahubKind.ACCOUNT,
            dst_peer=InfrahubKind.REFRESHTOKEN,
        )
        previous_rel = SchemaRelationshipInfo(
            name="coreaccount__internalrefreshtoken",
            branch_support=BranchSupportType.AGNOSTIC.value,
            src_peer=InfrahubKind.ACCOUNT,
            dst_peer=InfrahubKind.REFRESHTOKEN,
        )

        kwargs.pop("branch", None)

        super().__init__(new_rel=new_rel, previous_rel=previous_rel, branch=global_branch, **kwargs)


class Migration012RenameRelationshipThreadData(RelationshipDuplicateQuery):
    name = "migration_012_rename_rel_thread_data"
    type = QueryType.WRITE

    def __init__(self, **kwargs: Any):
        new_rel = SchemaRelationshipInfo(
            name="thread__account",
            branch_support=BranchSupportType.AGNOSTIC.value,
            src_peer=InfrahubKind.ACCOUNT,
            dst_peer=InfrahubKind.THREAD,
        )
        previous_rel = SchemaRelationshipInfo(
            name="coreaccount__corethread",
            branch_support=BranchSupportType.AGNOSTIC.value,
            src_peer=InfrahubKind.ACCOUNT,
            dst_peer=InfrahubKind.THREAD,
        )

        kwargs.pop("branch", None)

        super().__init__(new_rel=new_rel, previous_rel=previous_rel, branch=global_branch, **kwargs)


class Migration012RenameRelationshipCommentData(RelationshipDuplicateQuery):
    name = "migration_012_rename_rel_comment_data"
    type = QueryType.WRITE

    def __init__(self, **kwargs: Any):
        new_rel = SchemaRelationshipInfo(
            name="comment__account",
            branch_support=BranchSupportType.AGNOSTIC.value,
            src_peer=InfrahubKind.ACCOUNT,
            dst_peer=InfrahubKind.COMMENT,
        )
        previous_rel = SchemaRelationshipInfo(
            name="coreaccount__corecomment",
            branch_support=BranchSupportType.AGNOSTIC.value,
            src_peer=InfrahubKind.ACCOUNT,
            dst_peer=InfrahubKind.COMMENT,
        )

        kwargs.pop("branch", None)

        super().__init__(new_rel=new_rel, previous_rel=previous_rel, branch=default_branch, **kwargs)


class Migration012DeleteOldElementsSchema(DeleteElementInSchemaQuery):
    name = "migration_012_delete_old_elements_schema"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, **kwargs: Any):
        kwargs.pop("branch", None)

        super().__init__(
            element_names=["name", "password", "label", "description", "type", "role", "tokens"],
            node_name="Account",
            node_namespace="Core",
            branch=default_branch,
            **kwargs,
        )


class Migration012UpdateDisplayLabels(SchemaAttributeUpdateQuery):
    name = "migration_012_display_labels"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, **kwargs: Any):
        kwargs.pop("branch", None)

        super().__init__(
            attribute_name="display_labels",
            node_name="Account",
            node_namespace="Core",
            new_value="NULL",
            **kwargs,
        )


class Migration012UpdateOrderBy(SchemaAttributeUpdateQuery):
    name = "migration_012_order_by"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, **kwargs: Any):
        kwargs.pop("branch", None)

        super().__init__(
            attribute_name="order_by",
            node_name="Account",
            node_namespace="Core",
            new_value="NULL",
            **kwargs,
        )


class Migration012UpdateDefaultFilter(SchemaAttributeUpdateQuery):
    name = "migration_012_reset_default_filter"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, **kwargs: Any):
        kwargs.pop("branch", None)

        super().__init__(
            attribute_name="default_filter",
            node_name="Account",
            node_namespace="Core",
            new_value="NULL",
            **kwargs,
        )


class Migration012UpdateHFID(SchemaAttributeUpdateQuery):
    name = "migration_012_reset_hfid"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, **kwargs: Any):
        kwargs.pop("branch", None)

        super().__init__(
            attribute_name="human_friendly_id",
            node_name="Account",
            node_namespace="Core",
            new_value="NULL",
            **kwargs,
        )


class Migration012(GraphMigration):
    name: str = "012_convert_account_generic"
    queries: Sequence[type[Query]] = [
        Migration012DeleteOldElementsSchema,
        Migration012RenameTypeAttributeData,
        Migration012AddLabelData,
        Migration012RenameRelationshipAccountTokenData,
        Migration012RenameRelationshipRefreshTokenData,
        Migration012RenameRelationshipThreadData,
        Migration012RenameRelationshipCommentData,
        Migration012UpdateDefaultFilter,
        Migration012UpdateOrderBy,
        Migration012UpdateDisplayLabels,
        Migration012UpdateHFID,
    ]
    minimum_version: int = 11

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()

        return result



from collections import defaultdict
from dataclasses import asdict
from typing import AsyncGenerator

from infrahub.core.query import QueryType
from infrahub.database import InfrahubDatabase

from .models import VertexToAdd


class PatchPlanVertexAdder:
    def __init__(self, db: InfrahubDatabase, batch_size_limit: int = 1000) -> None:
        self.db = db
        self.batch_size_limit = batch_size_limit

    async def _run_add_query(self, labels: list[str], vertices_to_add: list[VertexToAdd]) -> dict[str, str]:
        labels_str = ":".join(labels)
        serial_vertices_to_add: list[dict[str, str | int | bool]] = [asdict(v) for v in vertices_to_add]
        query = """
UNWIND $vertices_to_add AS vertex_to_add
CREATE (v:%(labels)s)
SET v = vertex_to_add.after_props
RETURN vertex_to_add.identifier AS abstract_id, %(id_func_name)s(v) AS db_id
        """ % {
            "labels": labels_str,
            "id_func_name": self.db.get_id_function_name(),
        }
        # use transaction to make sure we record the results before committing them
        try:
            txn_db = self.db.start_transaction()
            async with txn_db as txn:
                results = await txn.execute_query(
                    query=query, params={"vertices_to_add": serial_vertices_to_add}, type=QueryType.WRITE
                )
                abstract_to_concrete_id_map: dict[str, str] = {}
                for result in results:
                    abstract_id = result.get("abstract_id")
                    concrete_id = result.get("db_id")
                    abstract_to_concrete_id_map[abstract_id] = concrete_id
        finally:
            await txn_db.close()
        return abstract_to_concrete_id_map

    async def execute(self, vertices_to_add: list[VertexToAdd]) -> AsyncGenerator[dict[str, str], None]:
        """
        Create vertices_to_add on the database.
        Returns a generator that yields dictionaries mapping VertexToAdd.identifier to the database-level ID of the newly created vertex.
        """
        vertices_map_queue: dict[frozenset[str], list[VertexToAdd]] = defaultdict(list)
        for vertex_to_add in vertices_to_add:
            frozen_labels = frozenset(vertex_to_add.labels)
            vertices_map_queue[frozen_labels].append(vertex_to_add)
            if len(vertices_map_queue[frozen_labels]) > self.batch_size_limit:
                yield await self._run_add_query(
                    labels=list(frozen_labels),
                    vertices_to_add=vertices_map_queue[frozen_labels],
                )
                vertices_map_queue[frozen_labels] = []

        for frozen_labels, vertices_group in vertices_map_queue.items():
            yield await self._run_add_query(labels=list(frozen_labels), vertices_to_add=vertices_group)



from __future__ import annotations

import base64
import hashlib
import hmac
from typing import TYPE_CHECKING, Any
from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict, Field, computed_field
from typing_extensions import Self

from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.core.timestamp import Timestamp
from infrahub.events.utils import get_all_infrahub_node_kind_events
from infrahub.git.repository import InfrahubReadOnlyRepository, InfrahubRepository
from infrahub.trigger.constants import NAME_SEPARATOR
from infrahub.trigger.models import EventTrigger, ExecuteWorkflow, TriggerDefinition, TriggerType
from infrahub.workflows.catalogue import WEBHOOK_PROCESS

if TYPE_CHECKING:
    from httpx import Response
    from infrahub_sdk.protocols import CoreCustomWebhook, CoreStandardWebhook, CoreTransformPython, CoreWebhook

    from infrahub.core.protocols import CoreWebhook as CoreWebhookNode
    from infrahub.services import InfrahubServices


class WebhookTriggerDefinition(TriggerDefinition):
    id: str
    type: TriggerType = TriggerType.WEBHOOK

    def generate_name(self) -> str:
        return f"{self.type.value}{NAME_SEPARATOR}{self.id}"

    @classmethod
    def generate_name_from_id(cls, id: str) -> str:
        return f"{TriggerType.WEBHOOK.value}{NAME_SEPARATOR}{id}"

    @classmethod
    def from_object(cls, obj: CoreWebhook | CoreWebhookNode) -> Self:
        event_trigger = EventTrigger()
        if obj.event_type.value == "all":
            event_trigger.events.add("infrahub.*")
        else:
            event_trigger.events.add(obj.event_type.value)

        if obj.branch_scope.value == "default_branch":
            event_trigger.match_related = {
                "prefect.resource.role": "infrahub.branch",
                "infrahub.resource.label": registry.default_branch,
            }
        elif obj.branch_scope.value == "other_branches":
            event_trigger.match_related = {
                "prefect.resource.role": "infrahub.branch",
                "infrahub.resource.label": f"!{registry.default_branch}",
            }

        if obj.node_kind.value and obj.event_type.value in get_all_infrahub_node_kind_events():
            event_trigger.match = {"infrahub.node.kind": obj.node_kind.value}

        definition = cls(
            id=obj.id,
            name=obj.name.value,
            trigger=event_trigger,
            actions=[
                ExecuteWorkflow(
                    workflow=WEBHOOK_PROCESS,
                    parameters={
                        "webhook_id": obj.id,
                        "webhook_name": obj.name.value,
                        "webhook_kind": obj.get_kind(),
                        "branch_name": "{{ event.resource['infrahub.branch.name'] }}",
                        "event_id": "{{ event.id }}",
                        "event_type": "{{ event.event }}",
                        "event_occured_at": "{{ event.occurred }}",
                        "event_payload": {
                            "__prefect_kind": "json",
                            "value": {"__prefect_kind": "jinja", "template": "{{ event.payload | tojson }}"},
                        },
                    },
                ),
            ],
        )

        return definition


class EventContext(BaseModel):
    id: str = Field(..., description="The internal id of the event")
    branch: str | None = Field(None, description="The branch associated with the event")
    account_id: str | None = Field(None, description="The id of the account associated with the event")
    occured_at: str = Field(..., description="The time when the event occurred")
    event: str = Field(..., description="The event type")

    @classmethod
    def from_event(cls, event_id: str, event_type: str, event_occured_at: str, event_payload: dict[str, Any]) -> Self:
        """Extract the context from the raw event we are getting from Prefect."""

        infrahub_context: dict[str, Any] = event_payload.get("context", {})
        account_info: dict[str, Any] = infrahub_context.get("account", {})
        branch_info: dict[str, Any] = infrahub_context.get("branch", {})

        return cls(
            id=event_id,
            branch=branch_info.get("name")
            if branch_info and branch_info.get("name") != registry.get_global_branch().name
            else None,
            account_id=account_info.get("account_id"),
            occured_at=event_occured_at,
            event=event_type,
        )


class Webhook(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    name: str = Field(...)
    url: str = Field(...)
    event_type: str = Field(...)
    validate_certificates: bool = Field(...)
    _payload: Any = None
    _headers: dict[str, Any] | None = None

    async def _prepare_payload(self, data: dict[str, Any], context: EventContext, service: InfrahubServices) -> None:  # noqa: ARG002
        self._payload = {"data": data, **context.model_dump()}

    def _assign_headers(self) -> None:
        self._headers = {}

    @computed_field  # type: ignore[prop-decorator]
    @property
    def webhook_type(self) -> str:
        return self.__class__.__name__

    async def prepare(self, data: dict[str, Any], context: EventContext, service: InfrahubServices) -> None:
        await self._prepare_payload(data=data, context=context, service=service)
        self._assign_headers()

    async def send(self, data: dict[str, Any], context: EventContext, service: InfrahubServices) -> Response:
        await self.prepare(data=data, context=context, service=service)
        return await service.http.post(url=self.url, json=self.get_payload(), headers=self._headers)

    def get_payload(self) -> dict[str, Any]:
        return self._payload

    def to_cache(self) -> dict[str, Any]:
        return self.model_dump()

    @classmethod
    def from_cache(cls, data: dict[str, Any]) -> Self:
        return cls(**data)


class CustomWebhook(Webhook):
    """Custom webhook"""

    @classmethod
    def from_object(cls, obj: CoreCustomWebhook) -> Self:
        return cls(
            name=obj.name.value,
            url=obj.url.value,
            event_type=obj.event_type.value,
            validate_certificates=obj.validate_certificates.value or False,
        )


class StandardWebhook(Webhook):
    shared_key: str = Field(...)

    def _assign_headers(self, uuid: UUID | None = None, at: Timestamp | None = None) -> None:
        message_id = f"msg_{uuid.hex}" if uuid else f"msg_{uuid4().hex}"
        timestamp = str(at.to_timestamp()) if at else str(Timestamp().to_timestamp())
        payload = self._payload or {}
        unsigned_data = f"{message_id}.{timestamp}.{payload}".encode()
        signature = self._sign(data=unsigned_data)

        self._headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "webhook-id": message_id,
            "webhook-timestamp": timestamp,
            "webhook-signature": f"v1,{base64.b64encode(signature).decode('utf-8')}",
        }

    def _sign(self, data: bytes) -> bytes:
        return hmac.new(key=self.shared_key.encode(), msg=data, digestmod=hashlib.sha256).digest()

    @classmethod
    def from_object(cls, obj: CoreStandardWebhook) -> Self:
        return cls(
            name=obj.name.value,
            url=obj.url.value,
            event_type=obj.event_type.value,
            validate_certificates=obj.validate_certificates.value or False,
            shared_key=obj.shared_key.value,
        )


class TransformWebhook(Webhook):
    repository_id: str = Field(...)
    repository_name: str = Field(...)
    repository_kind: str = Field(...)
    transform_name: str = Field(...)
    transform_class: str = Field(...)
    transform_file: str = Field(...)
    transform_timeout: int = Field(...)
    convert_query_response: bool = Field(...)

    async def _prepare_payload(self, data: dict[str, Any], context: EventContext, service: InfrahubServices) -> None:
        repo: InfrahubReadOnlyRepository | InfrahubRepository
        if self.repository_kind == InfrahubKind.READONLYREPOSITORY:
            repo = await InfrahubReadOnlyRepository.init(
                id=self.repository_id,
                name=self.repository_name,
                client=service.client,
                service=service,
            )
        else:
            repo = await InfrahubRepository.init(
                id=self.repository_id,
                name=self.repository_name,
                client=service.client,
                service=service,
            )

        branch = context.branch or repo.default_branch
        commit = repo.get_commit_value(branch_name=branch)

        self._payload = await repo.execute_python_transform.with_options(timeout_seconds=self.transform_timeout)(
            branch_name=branch,
            commit=commit,
            location=f"{self.transform_file}::{self.transform_class}",
            convert_query_response=self.convert_query_response,
            data={"data": data, **context.model_dump()},
            client=service.client,
        )  # type: ignore[misc]

    @classmethod
    def from_object(cls, obj: CoreCustomWebhook, transform: CoreTransformPython) -> Self:
        return cls(
            name=obj.name.value,
            url=obj.url.value,
            event_type=obj.event_type.value,
            validate_certificates=obj.validate_certificates.value or False,
            repository_id=transform.repository.id,
            repository_name=transform.repository.peer.name.value,
            repository_kind=transform.repository.peer.get_kind(),
            transform_name=transform.name.value,
            transform_class=transform.class_name.value,
            transform_file=transform.file_path.value,
            transform_timeout=transform.timeout.value,
            convert_query_response=transform.convert_query_response.value or False,
        )



from typing import Any

from infrahub.core.query import Query, QueryType
from infrahub.database import InfrahubDatabase


class EnrichedDiffDeleteQuery(Query):
    name = "enriched_diff_delete"
    type = QueryType.WRITE
    insert_return = False

    def __init__(self, enriched_diff_root_uuids: list[str], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.enriched_diff_root_uuids = enriched_diff_root_uuids

    async def query_init(self, db: InfrahubDatabase, **kwargs: Any) -> None:  # noqa: ARG002
        self.params = {"diff_root_uuids": self.enriched_diff_root_uuids}
        query = """
        MATCH (d_root:DiffRoot)
        WHERE d_root.uuid IN $diff_root_uuids
        OPTIONAL MATCH (d_root)-[*]->(diff_thing)
        DETACH DELETE diff_thing
        DETACH DELETE d_root
        """
        self.add_to_query(query=query)



from __future__ import annotations

import enum
from logging import Logger
from typing import TYPE_CHECKING, Any, Protocol, Union, runtime_checkable

from pydantic import BaseModel

if TYPE_CHECKING:
    import httpx


class HTTPMethod(str, enum.Enum):
    GET = "get"
    POST = "post"


class RequesterTransport(str, enum.Enum):
    HTTPX = "httpx"
    JSON = "json"


@runtime_checkable
class SyncRequester(Protocol):
    def __call__(
        self,
        url: str,
        method: HTTPMethod,
        headers: dict[str, Any],
        timeout: int,
        payload: dict | None = None,
    ) -> httpx.Response: ...


@runtime_checkable
class AsyncRequester(Protocol):
    async def __call__(
        self,
        url: str,
        method: HTTPMethod,
        headers: dict[str, Any],
        timeout: int,
        payload: dict | None = None,
    ) -> httpx.Response: ...


@runtime_checkable
class InfrahubLogger(Protocol):
    def debug(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """Send a debug event"""

    def info(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """Send an info event"""

    def warning(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """Send a warning event"""

    def error(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """Send an error event."""

    def critical(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """Send a critical event."""

    def exception(self, event: str | None = None, *args: Any, **kw: Any) -> Any:
        """Send an exception event."""


InfrahubLoggers = Union[InfrahubLogger, Logger]


class Order(BaseModel):
    disable: bool | None = None



from __future__ import annotations

from ..yaml import InfrahubFile, InfrahubFileKind
from .object import InfrahubObjectFileData, ObjectFile


class InfrahubMenuFileData(InfrahubObjectFileData):
    kind: str = "CoreMenuItem"

    @classmethod
    def enrich_node(cls, data: dict, context: dict) -> dict:
        if "kind" in data and "path" not in data:
            data["path"] = "/objects/" + data["kind"]

        if "list_index" in context and "order_weight" not in data:
            data["order_weight"] = (context["list_index"] + 1) * 1000

        return data


class MenuFile(ObjectFile):
    _spec: InfrahubMenuFileData | None = None

    @property
    def spec(self) -> InfrahubMenuFileData:
        if not self._spec:
            self._spec = InfrahubMenuFileData(**self.data.spec)
        return self._spec

    def validate_content(self) -> None:
        InfrahubFile.validate_content(self)
        if self.kind != InfrahubFileKind.MENU:
            raise ValueError("File is not an Infrahub Menu file")
        self._spec = InfrahubMenuFileData(**self.data.spec)



from __future__ import annotations

from enum import Enum


class DatabaseEdgeType(Enum):
    IS_PART_OF = "IS_PART_OF"
    HAS_ATTRIBUTE = "HAS_ATTRIBUTE"
    IS_RELATED = "IS_RELATED"
    HAS_VALUE = "HAS_VALUE"
    IS_VISIBLE = "IS_VISIBLE"
    IS_PROTECTED = "IS_PROTECTED"
    HAS_OWNER = "HAS_OWNER"
    HAS_SOURCE = "HAS_SOURCE"



from __future__ import annotations

import warnings
from collections.abc import MutableMapping
from enum import Enum
from typing import TYPE_CHECKING, Any, Union

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import Self

if TYPE_CHECKING:
    from ..node import InfrahubNode, InfrahubNodeSync

    InfrahubNodeTypes = Union[InfrahubNode, InfrahubNodeSync]


class RelationshipCardinality(str, Enum):
    ONE = "one"
    MANY = "many"


class BranchSupportType(str, Enum):
    AWARE = "aware"
    AGNOSTIC = "agnostic"
    LOCAL = "local"


class RelationshipKind(str, Enum):
    GENERIC = "Generic"
    ATTRIBUTE = "Attribute"
    COMPONENT = "Component"
    PARENT = "Parent"
    GROUP = "Group"
    HIERARCHY = "Hierarchy"
    PROFILE = "Profile"
    TEMPLATE = "Template"


class RelationshipDirection(str, Enum):
    BIDIR = "bidirectional"
    OUTBOUND = "outbound"
    INBOUND = "inbound"


class AttributeKind(str, Enum):
    ID = "ID"
    TEXT = "Text"
    STRING = "String"  # deprecated
    TEXTAREA = "TextArea"
    DATETIME = "DateTime"
    NUMBER = "Number"
    DROPDOWN = "Dropdown"
    EMAIL = "Email"
    PASSWORD = "Password"  # noqa: S105
    HASHEDPASSWORD = "HashedPassword"
    URL = "URL"
    FILE = "File"
    MAC_ADDRESS = "MacAddress"
    COLOR = "Color"
    BANDWIDTH = "Bandwidth"
    IPHOST = "IPHost"
    IPNETWORK = "IPNetwork"
    BOOLEAN = "Boolean"
    CHECKBOX = "Checkbox"
    LIST = "List"
    JSON = "JSON"
    ANY = "Any"

    def __getattr__(self, name: str) -> Any:
        if name == "STRING":
            warnings.warn(
                f"{name} is deprecated and will be removed in future versions.",
                DeprecationWarning,
                stacklevel=2,
            )
        return super().__getattribute__(name)


class SchemaState(str, Enum):
    PRESENT = "present"
    ABSENT = "absent"


class AllowOverrideType(str, Enum):
    NONE = "none"
    ANY = "any"


class RelationshipDeleteBehavior(str, Enum):
    NO_ACTION = "no-action"
    CASCADE = "cascade"


class AttributeSchema(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    id: str | None = None
    state: SchemaState = SchemaState.PRESENT
    name: str
    kind: AttributeKind
    label: str | None = None
    description: str | None = None
    default_value: Any | None = None
    unique: bool = False
    branch: BranchSupportType | None = None
    optional: bool = False
    choices: list[dict[str, Any]] | None = None
    enum: list[str | int] | None = None
    max_length: int | None = None
    min_length: int | None = None
    regex: str | None = None
    order_weight: int | None = None


class AttributeSchemaAPI(AttributeSchema):
    model_config = ConfigDict(use_enum_values=True)

    inherited: bool = False
    read_only: bool = False
    allow_override: AllowOverrideType = AllowOverrideType.ANY


class RelationshipSchema(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    id: str | None = None
    state: SchemaState = SchemaState.PRESENT
    name: str
    peer: str
    kind: RelationshipKind = RelationshipKind.GENERIC
    label: str | None = None
    description: str | None = None
    identifier: str | None = None
    min_count: int | None = None
    max_count: int | None = None
    direction: RelationshipDirection = RelationshipDirection.BIDIR
    on_delete: RelationshipDeleteBehavior | None = None
    cardinality: str = "many"
    branch: BranchSupportType | None = None
    optional: bool = True
    order_weight: int | None = None


class RelationshipSchemaAPI(RelationshipSchema):
    model_config = ConfigDict(use_enum_values=True)

    inherited: bool = False
    read_only: bool = False
    hierarchical: str | None = None
    allow_override: AllowOverrideType = AllowOverrideType.ANY


class BaseSchemaAttrRel(BaseModel):
    attributes: list[AttributeSchema] = Field(default_factory=list)
    relationships: list[RelationshipSchema] = Field(default_factory=list)


class BaseSchemaAttrRelAPI(BaseModel):
    attributes: list[AttributeSchemaAPI] = Field(default_factory=list)
    relationships: list[RelationshipSchemaAPI] = Field(default_factory=list)

    def get_field(self, name: str, raise_on_error: bool = True) -> AttributeSchemaAPI | RelationshipSchemaAPI | None:
        if attribute_field := self.get_attribute_or_none(name=name):
            return attribute_field

        if relationship_field := self.get_relationship_or_none(name=name):
            return relationship_field

        if not raise_on_error:
            return None

        raise ValueError(f"Unable to find the field {name}")

    def get_attribute(self, name: str) -> AttributeSchemaAPI:
        for item in self.attributes:
            if item.name == name:
                return item
        raise ValueError(f"Unable to find the attribute {name}")

    def get_attribute_or_none(self, name: str) -> AttributeSchemaAPI | None:
        for item in self.attributes:
            if item.name == name:
                return item
        return None

    def get_relationship(self, name: str) -> RelationshipSchemaAPI:
        for item in self.relationships:
            if item.name == name:
                return item
        raise ValueError(f"Unable to find the relationship {name}")

    def get_relationship_or_none(self, name: str) -> RelationshipSchemaAPI | None:
        for item in self.relationships:
            if item.name == name:
                return item
        return None

    def get_relationship_by_identifier(self, id: str, raise_on_error: bool = True) -> RelationshipSchemaAPI | None:
        for item in self.relationships:
            if item.identifier == id:
                return item

        if not raise_on_error:
            return None

        raise ValueError(f"Unable to find the relationship {id}")

    def get_matching_relationship(
        self, id: str, direction: RelationshipDirection = RelationshipDirection.BIDIR
    ) -> RelationshipSchemaAPI:
        valid_direction = RelationshipDirection.BIDIR
        if direction == RelationshipDirection.INBOUND:
            valid_direction = RelationshipDirection.OUTBOUND
        elif direction == RelationshipDirection.OUTBOUND:
            valid_direction = RelationshipDirection.INBOUND
        for item in self.relationships:
            if item.identifier == id and item.direction == valid_direction:
                return item
        raise ValueError(f"Unable to find the relationship {id} / ({valid_direction.value})")

    @property
    def attribute_names(self) -> list[str]:
        return [item.name for item in self.attributes]

    @property
    def relationship_names(self) -> list[str]:
        return [item.name for item in self.relationships]

    @property
    def mandatory_input_names(self) -> list[str]:
        return self.mandatory_attribute_names + self.mandatory_relationship_names

    @property
    def mandatory_attribute_names(self) -> list[str]:
        return [
            item.name
            for item in self.attributes
            if (not item.optional and item.default_value is None) and not item.read_only
        ]

    @property
    def mandatory_relationship_names(self) -> list[str]:
        return [item.name for item in self.relationships if not item.optional]

    @property
    def local_attributes(self) -> list[AttributeSchemaAPI]:
        return [item for item in self.attributes if not item.inherited]

    @property
    def local_relationships(self) -> list[RelationshipSchemaAPI]:
        return [item for item in self.relationships if not item.inherited]

    @property
    def unique_attributes(self) -> list[AttributeSchemaAPI]:
        return [item for item in self.attributes if item.unique]


class BaseSchema(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    id: str | None = None
    state: SchemaState = SchemaState.PRESENT
    name: str
    label: str | None = None
    namespace: str
    description: str | None = None
    include_in_menu: bool | None = None
    menu_placement: str | None = None
    display_labels: list[str] | None = None
    human_friendly_id: list[str] | None = None
    icon: str | None = None
    uniqueness_constraints: list[list[str]] | None = None
    documentation: str | None = None
    order_by: list[str] | None = None

    @property
    def kind(self) -> str:
        return self.namespace + self.name


class GenericSchema(BaseSchema, BaseSchemaAttrRel):
    def convert_api(self) -> GenericSchemaAPI:
        return GenericSchemaAPI(**self.model_dump())


class GenericSchemaAPI(BaseSchema, BaseSchemaAttrRelAPI):
    """A Generic can be either an Interface or a Union depending if there are some Attributes or Relationships defined."""

    hash: str | None = None
    used_by: list[str] = Field(default_factory=list)


class BaseNodeSchema(BaseSchema):
    model_config = ConfigDict(use_enum_values=True)

    inherit_from: list[str] = Field(default_factory=list)
    branch: BranchSupportType | None = None
    default_filter: str | None = None
    generate_profile: bool | None = None
    generate_template: bool | None = None
    parent: str | None = None
    children: str | None = None


class NodeSchema(BaseNodeSchema, BaseSchemaAttrRel):
    def convert_api(self) -> NodeSchemaAPI:
        return NodeSchemaAPI(**self.model_dump())


class NodeSchemaAPI(BaseNodeSchema, BaseSchemaAttrRelAPI):
    hash: str | None = None
    hierarchy: str | None = None


class ProfileSchemaAPI(BaseSchema, BaseSchemaAttrRelAPI):
    inherit_from: list[str] = Field(default_factory=list)


class TemplateSchemaAPI(BaseSchema, BaseSchemaAttrRelAPI):
    inherit_from: list[str] = Field(default_factory=list)


class NodeExtensionSchema(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    name: str | None = None
    kind: str
    description: str | None = None
    label: str | None = None
    inherit_from: list[str] = Field(default_factory=list)
    branch: BranchSupportType | None = None
    default_filter: str | None = None
    attributes: list[AttributeSchema] = Field(default_factory=list)
    relationships: list[RelationshipSchema] = Field(default_factory=list)


class SchemaRoot(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    version: str
    generics: list[GenericSchema] = Field(default_factory=list)
    nodes: list[NodeSchema] = Field(default_factory=list)
    node_extensions: list[NodeExtensionSchema] = Field(default_factory=list)

    def to_schema_dict(self) -> dict[str, Any]:
        return self.model_dump(exclude_unset=True, exclude_defaults=True)


class SchemaRootAPI(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    main: str | None = None
    generics: list[GenericSchemaAPI] = Field(default_factory=list)
    nodes: list[NodeSchemaAPI] = Field(default_factory=list)
    profiles: list[ProfileSchemaAPI] = Field(default_factory=list)
    templates: list[TemplateSchemaAPI] = Field(default_factory=list)


class BranchSchema(BaseModel):
    hash: str = Field(...)
    nodes: MutableMapping[str, GenericSchemaAPI | NodeSchemaAPI | ProfileSchemaAPI | TemplateSchemaAPI] = Field(
        default_factory=dict
    )

    @classmethod
    def from_api_response(cls, data: MutableMapping[str, Any]) -> Self:
        """
        Convert an API response from /api/schema into a BranchSchema object.
        """
        return cls.from_schema_root_api(data=SchemaRootAPI(**data))

    @classmethod
    def from_schema_root_api(cls, data: SchemaRootAPI) -> Self:
        """
        Convert a SchemaRootAPI object to a BranchSchema object.
        """
        nodes: MutableMapping[str, GenericSchemaAPI | NodeSchemaAPI | ProfileSchemaAPI | TemplateSchemaAPI] = {}
        for node in data.nodes:
            nodes[node.kind] = node

        for generic in data.generics:
            nodes[generic.kind] = generic

        for profile in data.profiles:
            nodes[profile.kind] = profile

        for template in data.templates:
            nodes[template.kind] = template

        schema_hash = data.main or ""

        return cls(hash=schema_hash, nodes=nodes)



from .models import ContextUnit, MeasurementDefinition

NODE_QUERY_TIME = MeasurementDefinition(
    name="node_query",
    description="Query some nodes",
    dimensions=["kind"],
    unit=ContextUnit.TIME,
)

NODE_CREATE_TIME = MeasurementDefinition(
    name="node_mutation_create",
    description="Create a new node",
    dimensions=["kind"],
    unit=ContextUnit.TIME,
)

NODE_UPDATE_TIME = MeasurementDefinition(
    name="node_mutation_update",
    description="Update an existing node",
    dimensions=["kind"],
    unit=ContextUnit.TIME,
)

NODE_DELETE_TIME = MeasurementDefinition(
    name="node_mutation_delete",
    description="Delete a node",
    dimensions=["kind"],
    unit=ContextUnit.TIME,
)

SCHEMA_INITIAL_LOAD_TIME = MeasurementDefinition(
    name="schema_initial_load",
    description="Load the initial schema",
    dimensions=["branch"],
    unit=ContextUnit.TIME,
)

SCHEMA_UPDATE_TIME = MeasurementDefinition(
    name="schema_update",
    description="Update the schema",
    dimensions=["branch"],
    unit=ContextUnit.TIME,
)

DATABASE_SIZE = MeasurementDefinition(
    name="database_size",
    description="Size of the database",
    unit=ContextUnit.DISK,
)

DIFF_CREATE_TIME = MeasurementDefinition(
    name="diff_create",
    description="Create a new diff",
    unit=ContextUnit.TIME,
)

DIFF_APPLY_TIME = MeasurementDefinition(
    name="diff_update",
    description="Update an existing diff",
    unit=ContextUnit.TIME,
)

BRANCH_CREATE_TIME = MeasurementDefinition(
    name="branch_create",
    description="Create a new branch",
    unit=ContextUnit.TIME,
)

BRANCH_MERGE_TIME = MeasurementDefinition(
    name="branch_merge",
    description="Merge a branch",
    unit=ContextUnit.TIME,
)

BRANCH_REBASE_TIME = MeasurementDefinition(
    name="branch_rebase",
    description="Rebase a branch",
    unit=ContextUnit.TIME,
)

SCRIPT_EXECUTION_TIME = MeasurementDefinition(
    name="script_execution",
    description="Execute a script",
    dimensions=["name"],
    unit=ContextUnit.TIME,
)

GENERATOR_EXECUTION_TIME = MeasurementDefinition(
    name="generator_execution",
    description="Execute a generator",
    dimensions=["name"],
    unit=ContextUnit.TIME,
)



from __future__ import annotations

from typing import (
    Any,
)

from typing_extensions import NotRequired, TypedDict


class NodeDiff(TypedDict):
    branch: str
    kind: str
    id: str
    action: str
    display_label: str
    elements: list[NodeDiffElement]


class NodeDiffElement(TypedDict):
    name: str
    element_type: str
    action: str
    summary: NodeDiffSummary
    peers: NotRequired[list[NodeDiffPeer]]


class NodeDiffSummary(TypedDict):
    added: int
    updated: int
    removed: int


class NodeDiffPeer(TypedDict):
    action: str
    summary: NodeDiffSummary


def get_diff_summary_query() -> str:
    return """
        query GetDiffTree($branch_name: String!) {
            DiffTree(branch: $branch_name) {
                nodes {
                    uuid
                    kind
                    status
                    label
                    num_added
                    num_updated
                    num_removed
                    attributes {
                        name
                        status
                        num_added
                        num_updated
                        num_removed
                    }
                    relationships {
                        name
                        status
                        cardinality
                        num_added
                        num_updated
                        num_removed
                        elements {
                            status
                            num_added
                            num_updated
                            num_removed
                        }
                    }
                }
            }
        }
    """


def diff_tree_node_to_node_diff(node_dict: dict[str, Any], branch_name: str) -> NodeDiff:
    element_diffs: list[NodeDiffElement] = []
    if "attributes" in node_dict:
        for attr_dict in node_dict["attributes"]:
            attr_diff = NodeDiffElement(
                action=str(attr_dict.get("status")),
                element_type="ATTRIBUTE",
                name=str(attr_dict.get("name")),
                summary={
                    "added": int(attr_dict.get("num_added") or 0),
                    "removed": int(attr_dict.get("num_removed") or 0),
                    "updated": int(attr_dict.get("num_updated") or 0),
                },
            )
            element_diffs.append(attr_diff)
    if "relationships" in node_dict:
        for relationship_dict in node_dict["relationships"]:
            is_cardinality_one = str(relationship_dict.get("cardinality")).upper() == "ONE"
            relationship_diff = NodeDiffElement(
                action=str(relationship_dict.get("status")),
                element_type="RELATIONSHIP_ONE" if is_cardinality_one else "RELATIONSHIP_MANY",
                name=str(relationship_dict.get("name")),
                summary={
                    "added": int(relationship_dict.get("num_added") or 0),
                    "removed": int(relationship_dict.get("num_removed") or 0),
                    "updated": int(relationship_dict.get("num_updated") or 0),
                },
            )
            if not is_cardinality_one and "elements" in relationship_dict:
                peer_diffs = []
                for element_dict in relationship_dict["elements"]:
                    peer_diffs.append(
                        NodeDiffPeer(
                            action=str(element_dict.get("status")),
                            summary={
                                "added": int(element_dict.get("num_added") or 0),
                                "removed": int(element_dict.get("num_removed") or 0),
                                "updated": int(element_dict.get("num_updated") or 0),
                            },
                        )
                    )
                relationship_diff["peers"] = peer_diffs
            element_diffs.append(relationship_diff)
    node_diff = NodeDiff(
        branch=branch_name,
        kind=str(node_dict.get("kind")),
        id=str(node_dict.get("uuid")),
        action=str(node_dict.get("action")),
        display_label=str(node_dict.get("label")),
        elements=element_diffs,
    )
    return node_diff



from typing import TYPE_CHECKING

from infrahub.core.branch import Branch
from infrahub.core.node import Node
from infrahub.core.node.constraints.grouped_uniqueness import NodeGroupedUniquenessConstraint
from infrahub.core.relationship.constraints.interface import RelationshipManagerConstraintInterface
from infrahub.database import InfrahubDatabase

if TYPE_CHECKING:
    from infrahub.core.relationship.model import RelationshipManager


class NodeConstraintRunner:
    def __init__(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        uniqueness_constraint: NodeGroupedUniquenessConstraint,
        relationship_manager_constraints: list[RelationshipManagerConstraintInterface],
    ) -> None:
        self.db = db
        self.branch = branch
        self.uniqueness_constraint = uniqueness_constraint
        self.relationship_manager_constraints = relationship_manager_constraints

    async def check(
        self, node: Node, field_filters: list[str] | None = None, skip_uniqueness_check: bool = False
    ) -> None:
        async with self.db.start_session() as db:
            await node.resolve_relationships(db=db)

            if not skip_uniqueness_check:
                await self.uniqueness_constraint.check(node, filters=field_filters)

            for relationship_name in node.get_schema().relationship_names:
                if field_filters and relationship_name not in field_filters:
                    continue
                relationship_manager: RelationshipManager = getattr(node, relationship_name)
                await relationship_manager.fetch_relationship_ids(db=db, force_refresh=True)
                for relationship_constraint in self.relationship_manager_constraints:
                    await relationship_constraint.check(relm=relationship_manager, node_schema=node.get_schema())



from abc import ABC, abstractmethod

from infrahub.core.schema import MainSchemaTypes

from ..model import RelationshipManager


class RelationshipManagerConstraintInterface(ABC):
    @abstractmethod
    async def check(self, relm: RelationshipManager, node_schema: MainSchemaTypes) -> None: ...



from infrahub_sdk.node import InfrahubNode
from prefect import task
from prefect.cache_policies import NONE

from infrahub import lock
from infrahub.artifacts.models import CheckArtifactCreate
from infrahub.core.constants import InfrahubKind
from infrahub.git.models import RequestArtifactGenerate
from infrahub.services import InfrahubServices


@task(name="define-artifact", task_run_name="Define Artifact", cache_policy=NONE)  # type: ignore[arg-type]
async def define_artifact(
    model: CheckArtifactCreate | RequestArtifactGenerate, service: InfrahubServices
) -> tuple[InfrahubNode, bool]:
    """Return an artifact together with a flag to indicate if the artifact is created now or already existed."""
    created = False
    if model.artifact_id:
        artifact = await service.client.get(kind=InfrahubKind.ARTIFACT, id=model.artifact_id, branch=model.branch_name)
    else:
        async with lock.registry.get(f"{model.target_id}-{model.artifact_definition}", namespace="artifact"):
            artifacts = await service.client.filters(
                kind=InfrahubKind.ARTIFACT,
                branch=model.branch_name,
                definition__ids=[model.artifact_definition],
                object__ids=[model.target_id],
            )
            if artifacts:
                artifact = artifacts[0]
            else:
                artifact = await service.client.create(
                    kind=InfrahubKind.ARTIFACT,
                    branch=model.branch_name,
                    data={
                        "name": model.artifact_name,
                        "status": "Pending",
                        "object": model.target_id,
                        "definition": model.artifact_definition,
                        "content_type": model.content_type,
                    },
                )
                await artifact.save(request_context=model.context.to_request_context())
                created = True
    return artifact, created



import ipaddress
from typing import TYPE_CHECKING

from prefect import flow

from infrahub.core import registry
from infrahub.core.ipam.reconciler import IpamReconciler
from infrahub.services import InfrahubServices
from infrahub.workflows.utils import add_branch_tag

from .model import IpamNodeDetails

if TYPE_CHECKING:
    from infrahub.core.ipam.constants import AllIPTypes


@flow(
    name="ipam_reconciliation",
    flow_run_name="branch-{branch}",
    description="Ensure the IPAM Tree is up to date",
    persist_result=False,
)
async def ipam_reconciliation(branch: str, ipam_node_details: list[IpamNodeDetails], service: InfrahubServices) -> None:
    async with service.database.start_session() as db:
        branch_obj = await registry.get_branch(db=db, branch=branch)

        await add_branch_tag(branch_name=branch_obj.name)

        ipam_reconciler = IpamReconciler(db=db, branch=branch_obj)

        for ipam_node_detail_item in ipam_node_details:
            if ipam_node_detail_item.is_address:
                ip_value: AllIPTypes = ipaddress.ip_interface(ipam_node_detail_item.ip_value)
            else:
                ip_value = ipaddress.ip_network(ipam_node_detail_item.ip_value)
            await ipam_reconciler.reconcile(
                ip_value=ip_value,
                namespace=ipam_node_detail_item.namespace_id,
                node_uuid=ipam_node_detail_item.node_uuid,
                is_delete=ipam_node_detail_item.is_delete,
            )



from infrahub.core.constants import (
    BranchSupportType,
    InfrahubKind,
)
from infrahub.core.constants import RelationshipCardinality as Cardinality
from infrahub.core.constants import RelationshipKind as RelKind

from ...attribute_schema import AttributeSchema as Attr
from ...generic_schema import GenericSchema
from ...node_schema import NodeSchema
from ...relationship_schema import (
    RelationshipSchema as Rel,
)

core_group = GenericSchema(
    name="Group",
    namespace="Core",
    description="Generic Group Object.",
    label="Group",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["label__value"],
    include_in_menu=False,
    icon="mdi:group",
    hierarchical=True,
    branch=BranchSupportType.AWARE,
    uniqueness_constraints=[["name__value"]],
    attributes=[
        Attr(name="name", kind="Text", unique=True),
        Attr(name="label", kind="Text", optional=True),
        Attr(name="description", kind="Text", optional=True),
        Attr(name="group_type", kind="Text", enum=["default", "internal"], default_value="default", optional=False),
    ],
    relationships=[
        Rel(
            name="members",
            peer=InfrahubKind.NODE,
            optional=True,
            identifier="group_member",
            cardinality=Cardinality.MANY,
            branch=BranchSupportType.AWARE,
        ),
        Rel(
            name="subscribers",
            peer=InfrahubKind.NODE,
            optional=True,
            identifier="group_subscriber",
            cardinality=Cardinality.MANY,
        ),
    ],
)

core_standard_group = NodeSchema(
    name="StandardGroup",
    namespace="Core",
    description="Group of nodes of any kind.",
    include_in_menu=False,
    icon="mdi:account-group",
    label="Standard Group",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    branch=BranchSupportType.AWARE,
    inherit_from=[InfrahubKind.GENERICGROUP],
    generate_profile=False,
)

core_generator_group = NodeSchema(
    name="GeneratorGroup",
    namespace="Core",
    description="Group of nodes that are created by a generator.",
    include_in_menu=False,
    icon="mdi:state-machine",
    label="Generator Group",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    branch=BranchSupportType.LOCAL,
    inherit_from=[InfrahubKind.GENERICGROUP],
    generate_profile=False,
)

core_graphql_query_group = NodeSchema(
    name="GraphQLQueryGroup",
    namespace="Core",
    description="Group of nodes associated with a given GraphQLQuery.",
    include_in_menu=False,
    icon="mdi:account-group",
    label="GraphQL Query Group",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    branch=BranchSupportType.LOCAL,
    inherit_from=[InfrahubKind.GENERICGROUP],
    generate_profile=False,
    attributes=[
        Attr(name="parameters", kind="JSON", optional=True),
    ],
    relationships=[
        Rel(
            name="query",
            peer=InfrahubKind.GRAPHQLQUERY,
            optional=False,
            cardinality=Cardinality.ONE,
            kind=RelKind.ATTRIBUTE,
        ),
    ],
)



import re

from pydantic import field_validator

from infrahub.core.models import HashableModel

HTML_COLOR = re.compile(r"#[0-9a-fA-F]{6}\b")


class DropdownChoice(HashableModel):
    name: str
    description: str | None = None
    color: str | None = None
    label: str | None = None

    _sort_by: list[str] = ["name"]

    @field_validator("color")
    @classmethod
    def kind_options(cls, v: str) -> str:
        if not v:
            return v
        if isinstance(v, str) and HTML_COLOR.match(v):
            return v.lower()

        raise ValueError("Color must be a valid HTML color code")



from infrahub.core.constants import PathType
from infrahub.core.path import DataPath
from infrahub.database import InfrahubDatabase
from infrahub.log import get_logger

from ..model.path import CalculatedDiffs, EnrichedDiffRoot
from .interface import DiffEnricherInterface

log = get_logger()


class DiffPathIdentifierEnricher(DiffEnricherInterface):
    """Add path identifiers to every element in the diff"""

    def __init__(self, db: InfrahubDatabase) -> None:
        self.db = db
        self._diff_branch_name: str | None = None

    @property
    def diff_branch_name(self) -> str:
        if not self._diff_branch_name:
            raise RuntimeError("diff_branch_name not set")
        return self._diff_branch_name

    async def enrich(self, enriched_diff_root: EnrichedDiffRoot, calculated_diffs: CalculatedDiffs) -> None:  # noqa: ARG002
        self._diff_branch_name = enriched_diff_root.diff_branch_name
        for node in enriched_diff_root.nodes:
            node_path = DataPath(
                branch=enriched_diff_root.diff_branch_name,
                path_type=PathType.NODE,
                node_id=node.uuid,
                kind=node.kind,
            )
            node.path_identifier = node_path.get_path()
            for attribute in node.attributes:
                attribute_path = DataPath(
                    branch=enriched_diff_root.diff_branch_name,
                    path_type=PathType.ATTRIBUTE,
                    node_id=node.uuid,
                    kind=node.kind,
                    field_name=attribute.name,
                )
                attribute.path_identifier = attribute_path.get_path()
                for attribute_property in attribute.properties:
                    property_path = attribute_path.model_copy()
                    property_path.property_name = attribute_property.property_type.value
                    attribute_property.path_identifier = property_path.get_path()
            if not node.relationships:
                continue
            for relationship in node.relationships:
                path_type = PathType.from_relationship(relationship.cardinality)
                relationship_path = DataPath(
                    branch=enriched_diff_root.diff_branch_name,
                    path_type=path_type,
                    node_id=node.uuid,
                    kind=node.kind,
                    field_name=relationship.name,
                )
                relationship.path_identifier = relationship_path.get_path()
                for relationship_element in relationship.relationships:
                    relationship_element_path = relationship_path.model_copy()
                    relationship_element_path.peer_id = relationship_element.peer_id
                    relationship_element.path_identifier = relationship_element_path.get_path(with_peer=False)
                    for relationship_property in relationship_element.properties:
                        relationship_property_path = relationship_element_path.model_copy()
                        relationship_property_path.property_name = relationship_property.property_type.value
                        relationship_property.path_identifier = relationship_property_path.get_path(with_peer=False)
        log.info("Path identifier diff enrichment complete.")
